"""
Project: BlueMath_tk
Sub-Module: datamining
Author: GeoOcean Research Group, Universidad de Cantabria
Creation Date: 19 January 2024
License: MIT
Repository: https://github.com/GeoOcean/BlueMath_tk.git
Status: Under development (Working)
"""

# Import essential functions/classes to be available at the package level.
from ._base_datamining import (
    BaseClustering,
    BaseReduction,
    BaseSampling,
    ClusteringComparator,
)
from .kma import KMA
from .lhs import LHS
from .mda import MDA
from .pca import PCA
from .som import SOM

# Optionally, define the module's `__all__` variable to control what gets imported when using `from module import *`.
__all__ = [
    "BaseSampling",
    "BaseClustering",
    "BaseReduction",
    "ClusteringComparator",
    "LHS",
    "MDA",
    "KMA",
    "SOM",
    "PCA",
]



import os.path as op

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from hydromt_sfincs import SfincsModel, utils

from bluemath_tk.wrappers.sfincs.sfincs_wrapper import SfincsModelWrapper


class SfincsPabloModelWrapper(SfincsModelWrapper):
    """
    Wrapper for the SFINCS model (Pablo version 26/02/2025).
    """

    p_data = "/path/to/data"
    p_hybeat = "/path/to/hybeat"
    cluster_centroids = pd.read_csv(op.join(p_data, "tc_representative.csv"))
    simulation_xr = xr.open_dataset(op.join(p_data, "representative_events.nc"))
    location_points = pd.read_csv(op.join(p_data, "forcing_points.csv"))
    datasets_dep = [{"elevtn": op.join(p_data, "apia_5m_soften.tif")}]
    catalogs_list = [
        op.join(p_data, "catalogues/manning_cat.yml"),
        op.join(p_data, "catalogues/topo_cat.yml"),
    ]

    def build_case(
        self,
        case_context: dict,
        case_dir: str,
    ) -> None:
        """
        Build the input files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        # Load TC position from cluster centroids
        TC_pos = self.cluster_centroids.TC_pos.values[int(case_context.get("case_num"))]

        # Instantiate the SFINCS model
        sf_model = SfincsModel(data_libs=self.catalogs_list, root=case_dir, mode="w+")

        # Set different parameters for the model
        sf_model.setup_grid(
            x0=408002.5,
            y0=8467002.5,
            dx=5.0,
            dy=5.0,
            nmax=1523,
            mmax=2494,
            rotation=0,
            epsg=32702,
        )
        sf_model.setup_dep(datasets_dep=self.datasets_dep)
        sf_model.setup_mask_active(zmin=-2, zmax=10, reset_mask=True)
        sf_model.setup_mask_bounds(btype="waterlevel", zmax=-1.99, reset_bounds=True)
        sf_model.setup_mask_bounds(btype="outflow", zmin=1, reset_bounds=False)
        sf_model.setup_config(
            **{
                "tref": "20000101 000000",
                "tstart": "20000101 000000",
                "tstop": "20000101 060000",
                "dtout": "100",
            }
        )

        # Load wave data for the selected TC
        wave_setup = pd.read_csv(
            op.join(self.p_hybeat, f"tc_{TC_pos}/HyBeat_level_tc_{TC_pos}.csv"),
            index_col=0,
        )
        wave_ig = pd.read_csv(
            op.join(self.p_hybeat, f"tc_{TC_pos}/HyBeat_ig_tc_{TC_pos}.csv"),
            index_col=0,
        )
        swl = pd.read_csv(
            op.join(self.p_hybeat, f"tc_{TC_pos}/HyBeat_swl_tc_{TC_pos}.csv"),
            index_col=0,
        )

        # Create custom waterlevel forcing depending on swl values
        if np.sum(swl.iloc[9:16]["SWL"].values) > 0:
            wave_setup = wave_setup.iloc[9:16]
            wave_ig = wave_ig.iloc[9:16]
            swl = swl.iloc[9:16]
            t_10s = np.arange(0, 6 * 3600, 10)
            transects = np.arange(60, 110, 1)
            waterlevels = pd.DataFrame()
            hours = np.arange(0, 7 * 3600, 3600)
            for i in transects:
                setup = np.array(wave_setup[f"{i}.0"].values)
                ig = np.array(wave_ig[f"{i}.0"].values)
                wave_setup_interp = np.interp(t_10s, hours, setup)
                swl_interp = np.interp(t_10s, hours, swl["SWL"].values)
                ig_amplitude_interp = np.interp(t_10s, hours, ig)
                ig_interp = ig_amplitude_interp * np.sin(2 * np.pi * t_10s / 200)
                _result = np.column_stack(
                    (t_10s, wave_setup_interp, swl_interp, ig_interp)
                )
                twl = wave_setup_interp + swl_interp + ig_interp
                waterlevels[f"{i}.0"] = twl

            waterlevels.columns = range(1, len(waterlevels.columns) + 1)
            # Loc points
            x = self.location_points.x.values
            y = self.location_points.y.values

            # Add to Geopandas dataframe as needed by HydroMT
            pnts = gpd.points_from_xy(x, y)
            index = np.arange(
                1, len(x) + 1, 1
            )  # NOTE that the index should start at one
            bnd = gpd.GeoDataFrame(index=index, geometry=pnts, crs=sf_model.crs)

            # Add time series
            time = pd.date_range(
                start=utils.parse_datetime(sf_model.config["tstart"]),
                end=utils.parse_datetime(sf_model.config["tstop"]),
                periods=len(waterlevels),
            )
            waterlevels.index = time
            sf_model.setup_waterlevel_forcing(timeseries=waterlevels, locations=bnd)
            sf_model.write()  # write all

        else:
            swl_pos = np.nonzero(swl["SWL"].values)[0]
            sf_model.setup_config(
                **{
                    "tref": "20000101 000000",
                    "tstart": "20000101 000000",
                    "tstop": f"20000101 {len(swl_pos):02}0000",
                    "dtout": "100",
                }
            )

            wave_setup = wave_setup.iloc[swl_pos]
            wave_ig = wave_ig.iloc[swl_pos]
            swl = swl.iloc[swl_pos]
            t_10s = np.arange(0, len(swl_pos) * 3600, 10)
            transects = np.arange(60, 110, 1)
            waterlevels = pd.DataFrame()
            hours = np.arange(0, (len(swl_pos)) * 3600, 3600)

            for i in transects:
                setup = np.array(wave_setup[f"{i}.0"].values)
                ig = np.array(wave_ig[f"{i}.0"].values)
                wave_setup_interp = np.interp(t_10s, hours, setup)
                swl_interp = np.interp(t_10s, hours, swl["SWL"].values)
                ig_amplitude_interp = np.interp(t_10s, hours, ig)
                ig_interp = ig_amplitude_interp * np.sin(2 * np.pi * t_10s / 200)
                _result = np.column_stack(
                    (t_10s, wave_setup_interp, swl_interp, ig_interp)
                )
                twl = wave_setup_interp + swl_interp + ig_interp
                waterlevels[f"{i}.0"] = twl

            waterlevels.columns = range(1, len(waterlevels.columns) + 1)
            # Loc points
            x = self.location_points.x.values
            y = self.location_points.y.values

            # Add to Geopandas dataframe as needed by HydroMT
            pnts = gpd.points_from_xy(x, y)
            index = np.arange(
                1, len(x) + 1, 1
            )  # NOTE that the index should start at one
            bnd = gpd.GeoDataFrame(index=index, geometry=pnts, crs=sf_model.crs)

            # Add time series
            time = pd.date_range(
                start=utils.parse_datetime(sf_model.config["tstart"]),
                end=utils.parse_datetime(sf_model.config["tstop"]),
                periods=len(waterlevels),
            )
            waterlevels.index = time
            sf_model.setup_waterlevel_forcing(timeseries=waterlevels, locations=bnd)
            sf_model.write()  # write all

    def build_cases(
        self,
        mode: str = "one_by_one",
    ) -> None:
        """
        Build the input files for all cases.

        Parameters
        ----------
        catalogs_list : List[str]
            The list of catalogs.
        mode : str, optional
            The mode to build the cases. Default is "one_by_one".

        Raises
        ------
        ValueError
            If the cases were not properly built
        """

        super().build_cases(mode=mode)
        if not self.cases_context or not self.cases_dirs:
            raise ValueError("Cases were not properly built.")
        for case_context, case_dir in zip(self.cases_context, self.cases_dirs):
            self.build_case(
                case_context=case_context,
                case_dir=case_dir,
            )


# Usage example
if __name__ == "__main__":
    # Define the input parameters
    templates_dir = (
        "/home/tausiaj/GitHub-GeoOcean/BlueMath/bluemath_tk/wrappers/sfincs/templates/"
    )
    output_dir = "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_cases/sfincs/pablo/"
    # Load swan model parameters
    model_parameters = {"waterlevel": [5.0, 20.0, 50.0, 100.0]}
    # Create an instance of the SWAN model wrapper
    sfincs_wrapper = SfincsPabloModelWrapper(
        templates_dir=templates_dir,
        model_parameters=model_parameters,
        output_dir=output_dir,
    )
    # Build the input files
    sfincs_wrapper.build_cases()
    # List available launchers
    print(sfincs_wrapper.list_available_launchers())
    # Run the model
    sfincs_wrapper.run_cases(launcher="docker", parallel=True)



from typing import List

import pandas as pd
from scipy.stats import qmc

from ..core.decorators import validate_data_lhs
from ._base_datamining import BaseSampling


class LHSError(Exception):
    """
    Custom exception for LHS class.
    """

    def __init__(self, message: str = "LHS error occurred."):
        self.message = message
        super().__init__(self.message)


class LHS(BaseSampling):
    """
    Latin Hypercube Sampling (LHS) class.

    This class performs the LHS algorithm for some input data.

    Attributes
    ----------
    num_dimensions : int
        The number of dimensions to use in the LHS algorithm.
    seed : int
        The random seed to use.
    lhs : qdc.LatinHypercube
        The Latin Hypercube object.
    data : pd.DataFrame
        The LHS samples dataframe.

    Methods
    -------
    generate(dimensions_names, lower_bounds, upper_bounds, num_samples)
        Generate LHS samples.

    Notes
    -----
    - This class is designed to perform the LHS algorithm.

    Examples
    --------
    >>> from bluemath_tk.datamining.lhs import LHS
    >>> dimensions_names = ['CM', 'SS', 'Qb']
    >>> lower_bounds = [0.5, -0.2, 1]
    >>> upper_bounds = [5.3, 1.5, 200]
    >>> lhs = LHS(num_dimensions=3, seed=0)
    >>> lhs_sampled_df = lhs.generate(
    ...     dimensions_names=dimensions_names,
    ...     lower_bounds=lower_bounds,
    ...     upper_bounds=upper_bounds,
    ...     num_samples=100,
    ... )
    """

    def __init__(self, num_dimensions: int, seed: int = 1) -> None:
        """
        Initializes the LHS class.

        Parameters
        ----------
        num_dimensions : int
            The number of dimensions to use in the LHS algorithm.
            Must be greater than 0.
        seed : int, optional
            The random seed to use.
            Must be greater or equal to 0.
            Default to 1.

        Raises
        ------
        ValueError
            If num_dimensions or num_samples is not greater than 0.
            Or if seed is not greater or equal to 0.
        """

        super().__init__()
        self.set_logger_name(name=self.__class__.__name__)
        if num_dimensions > 0:
            self.num_dimensions = int(num_dimensions)
        else:
            raise ValueError("Variable num_dimensions must be > 0")
        if seed >= 0:
            self.seed = int(seed)
        else:
            raise ValueError("Variable seed must be >= 0")
        self._lhs: qmc.LatinHypercube = qmc.LatinHypercube(
            d=self.num_dimensions, seed=self.seed
        )
        self._data: pd.DataFrame = pd.DataFrame()

    @property
    def lhs(self) -> qmc.LatinHypercube:
        return self._lhs

    @property
    def data(self) -> pd.DataFrame:
        return self._data

    @validate_data_lhs
    def generate(
        self,
        dimensions_names: List[str],
        lower_bounds: List[float],
        upper_bounds: List[float],
        num_samples: int,
    ) -> pd.DataFrame:
        """
        Generate LHS samples.

        Parameters
        ----------
        dimensions_names : List[str]
            The names of the dimensions.
        lower_bounds : List[float]
            The lower bounds of the dimensions.
        upper_bounds : List[float]
            The upper bounds of the dimensions.
        num_samples : int
            The number of samples to generate.
            Must be greater than 0.

        Returns
        -------
        self.data : pd.DataFrame
            The LHS samples.
        """

        lhs_samples = self.lhs.random(n=num_samples)
        lhs_scaled_data = qmc.scale(
            sample=lhs_samples, l_bounds=lower_bounds, u_bounds=upper_bounds
        )
        self._data = pd.DataFrame(data=lhs_scaled_data, columns=dimensions_names)
        return self.data



import importlib
import logging
import os
import pickle
import sys
from abc import ABC, abstractmethod
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from typing import Any, Callable, List, Tuple, TypeVar, Union

import numpy as np
import pandas as pd
import xarray as xr
from sklearn.metrics import (
    explained_variance_score,
    mean_absolute_error,
    mean_squared_error,
    r2_score,
)
from sklearn.preprocessing import StandardScaler

from .logging import get_file_logger
from .operations import (
    denormalize,
    destandarize,
    get_degrees_from_uv,
    get_uv_components,
    normalize,
    standarize,
)

T = TypeVar("T")


class BlueMathModel(ABC):
    """
    Abstract base class for handling default functionalities across the project.
    """

    @abstractmethod
    def __init__(self) -> None:
        self._logger: logging.Logger = None
        self._exclude_attributes: List[str] = []
        self.num_workers: int = 1

        # [UNDER DEVELOPMENT] Below, we try to generalise parallel processing
        bluemath_num_workers = os.environ.get("BLUEMATH_NUM_WORKERS", None)
        omp_num_threads = os.environ.get("OMP_NUM_THREADS", None)
        if bluemath_num_workers is not None:
            self.logger.warning(
                f"Setting self.num_workers to {bluemath_num_workers} due to BLUEMATH_NUM_WORKERS. \n"
                "Change it using self.set_num_processors_to_use method. \n"
                "Also setting OMP_NUM_THREADS to 1, to avoid conflicts with BlueMath parallel processing."
            )
            self.set_num_processors_to_use(num_processors=int(bluemath_num_workers))
            self.set_omp_num_threads(num_threads=1)
        elif omp_num_threads is not None:
            self.logger.warning(
                f"Changing variable OMP_NUM_THREADS from {omp_num_threads} to 1. \n"
                f"And setting self.num_workers to {omp_num_threads}. \n"
                "To avoid conflicts with BlueMath parallel processing."
            )
            self.set_omp_num_threads(num_threads=1)
            self.set_num_processors_to_use(num_processors=int(omp_num_threads))
        else:
            self.num_workers = 1  # self.get_num_processors_available()
            self.logger.warning(
                f"Setting self.num_workers to {self.num_workers}. "
                "Change it using self.set_num_processors_to_use method."
            )

    def __getstate__(self):
        """Exclude certain attributes from being pickled."""

        state = self.__dict__.copy()
        for attr in self._exclude_attributes:
            if attr in state:
                del state[attr]
        # Iterate through the state attributes, warning about xr.Datasets
        for key, value in state.items():
            if isinstance(value, xr.Dataset) or isinstance(value, xr.DataArray):
                self.logger.warning(
                    f"Attribute {key} is an xarray Dataset / Dataarray and will be pickled!"
                )

        return state

    @property
    def logger(self) -> logging.Logger:
        if self._logger is None:
            self._logger = get_file_logger(name=self.__class__.__name__)
        return self._logger

    @logger.setter
    def logger(self, value: logging.Logger) -> None:
        self._logger = value

    def set_logger_name(
        self, name: str, level: str = "INFO", console: bool = True
    ) -> None:
        """Sets the name of the logger."""

        self.logger = get_file_logger(name=name, level=level, console=console)

    def save_model(self, model_path: str, exclude_attributes: List[str] = None) -> None:
        """Saves the model to a file."""

        self.logger.info(f"Saving model to {model_path}")
        if exclude_attributes is not None:
            self._exclude_attributes += exclude_attributes
        with open(model_path, "wb") as f:
            pickle.dump(self, f)

    def load_model(self, model_path: str) -> "BlueMathModel":
        """Loads the model from a file."""

        self.logger.info(f"Loading model from {model_path}")
        with open(model_path, "rb") as f:
            model = pickle.load(f)

        return model

    def list_class_attributes(self) -> list:
        """
        Lists the attributes of the class.

        Returns
        -------
        list
            The attributes of the class.
        """

        return [
            attr
            for attr in dir(self)
            if not callable(getattr(self, attr)) and not attr.startswith("__")
        ]

    def list_class_methods(self) -> list:
        """
        Lists the methods of the class.

        Returns
        -------
        list
            The methods of the class.
        """

        return [
            attr
            for attr in dir(self)
            if callable(getattr(self, attr)) and not attr.startswith("__")
        ]

    def check_nans(
        self,
        data: Union[np.ndarray, pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset],
        replace_value: Union[float, callable] = None,
        raise_error: bool = False,
    ) -> Union[np.ndarray, pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset]:
        """
        Checks for NaNs in the data and optionally replaces them.

        Parameters
        ----------
        data : np.ndarray, pd.Series, pd.DataFrame, xr.DataArray or xr.Dataset
            The data to check for NaNs.
        replace_value : float or callable, optional
            The value to replace NaNs with. If None, NaNs will not be replaced.
            If a callable is provided, it will be called and the result will be returned.
            Default is None.
        raise_error : bool, optional
            Whether to raise an error if NaNs are found. Default is False.

        Returns
        -------
        data : np.ndarray, pd.Series, pd.DataFrame, xr.DataArray or xr.Dataset
            The data with NaNs optionally replaced.

        Raises
        ------
        ValueError
            If NaNs are found and raise_error is True.

        Notes
        -----
        - This method is intended to be used in classes that inherit from the BlueMathModel class.
        - The method checks for NaNs in the data and optionally replaces them with the specified value.
        """

        # If replace_value is a callable, just call and return it
        if callable(replace_value):
            self.logger.debug(f"Replace value is a callable. Calling {replace_value}.")
            return replace_value(data)
        if isinstance(data, np.ndarray):
            if np.isnan(data).any():
                if raise_error:
                    raise ValueError("Data contains NaNs.")
                self.logger.warning("Data contains NaNs.")
                if replace_value is not None:
                    data = np.nan_to_num(data, nan=replace_value)
                    self.logger.info(f"NaNs replaced with {replace_value}.")
        elif isinstance(data, pd.Series) or isinstance(data, pd.DataFrame):
            if data.isnull().values.any():
                if raise_error:
                    raise ValueError("Data contains NaNs.")
                self.logger.warning("Data contains NaNs.")
                if replace_value is not None:
                    data.fillna(replace_value, inplace=True)
                    self.logger.info(f"NaNs replaced with {replace_value}.")
        elif isinstance(data, xr.DataArray) or isinstance(data, xr.Dataset):
            if data.isnull().any():
                if raise_error:
                    raise ValueError("Data contains NaNs.")
                self.logger.warning("Data contains NaNs.")
                if replace_value is not None:
                    data = data.fillna(replace_value)
                    self.logger.info(f"NaNs replaced with {replace_value}.")
        else:
            self.logger.warning("Data type not supported for NaN check.")

        return data

    def normalize(
        self, data: Union[pd.DataFrame, xr.Dataset], custom_scale_factor: dict = {}
    ) -> Tuple[Union[pd.DataFrame, xr.Dataset], dict]:
        """
        Normalize data to 0-1 using min max scaler approach.
        More info in bluemath_tk.core.operations.normalize.

        Parameters
        ----------
        data : pd.DataFrame or xr.Dataset
            The data to normalize.
        custom_scale_factor : dict, optional
            Custom scale factors for normalization.

        Returns
        -------
        normalized_data : pd.DataFrame or xr.Dataset
            The normalized data.
        scale_factor : dict
            The scale factors used for normalization.
        """

        normalized_data, scale_factor = normalize(
            data=data, custom_scale_factor=custom_scale_factor, logger=self.logger
        )
        return normalized_data, scale_factor

    def denormalize(
        self, normalized_data: pd.DataFrame, scale_factor: dict
    ) -> pd.DataFrame:
        """
        Denormalize data using provided scale_factor.
        More info in bluemath_tk.core.operations.denormalize.

        Parameters
        ----------
        normalized_data : pd.DataFrame
            The normalized data to denormalize.
        scale_factor : dict
            The scale factors used for denormalization.

        Returns
        -------
        data : pd.DataFrame
            The denormalized data.
        """

        data = denormalize(normalized_data=normalized_data, scale_factor=scale_factor)
        return data

    def standarize(
        self,
        data: Union[np.ndarray, pd.DataFrame, xr.Dataset],
        scaler: StandardScaler = None,
        transform: bool = False,
    ) -> Tuple[Union[np.ndarray, pd.DataFrame, xr.Dataset], StandardScaler]:
        """
        Standarize data using StandardScaler.
        More info in bluemath_tk.core.operations.standarize.

        Parameters
        ----------
        data : np.ndarray, pd.DataFrame or xr.Dataset
            Input data to be standarized.
        scaler : StandardScaler, optional
            Scaler object to use for standarization. Default is None.
        transform : bool
            Whether to just transform the data. Default to False.

        Returns
        -------
        standarized_data : np.ndarray, pd.DataFrame or xr.Dataset
            Standarized data.
        scaler : StandardScaler
            Scaler object used for standarization.
        """

        standarized_data, scaler = standarize(
            data=data, scaler=scaler, transform=transform
        )
        return standarized_data, scaler

    def destandarize(
        self,
        standarized_data: Union[np.ndarray, pd.DataFrame, xr.Dataset],
        scaler: StandardScaler,
    ) -> Union[np.ndarray, pd.DataFrame, xr.Dataset]:
        """
        Destandarize data using provided scaler.
        More info in bluemath_tk.core.operations.destandarize.

        Parameters
        ----------
        standarized_data : np.ndarray, pd.DataFrame or xr.Dataset
            Standarized data to be destandarized.
        scaler : StandardScaler
            Scaler object used for standarization.

        Returns
        -------
        data : np.ndarray, pd.DataFrame or xr.Dataset
            Destandarized data.
        """

        data = destandarize(standarized_data=standarized_data, scaler=scaler)
        return data

    @staticmethod
    def get_metrics(
        data1: Union[pd.DataFrame, xr.Dataset],
        data2: Union[pd.DataFrame, xr.Dataset],
    ) -> pd.DataFrame:
        """
        Gets the metrics of the model.

        Parameters
        ----------
        data1 : pd.DataFrame or xr.Dataset
            The first dataset.
        data2 : pd.DataFrame or xr.Dataset
            The second dataset.

        Returns
        -------
        metrics : pd.DataFrame
            The metrics of the model.

        Raises
        ------
        ValueError
            If the DataFrames or Datasets have different shapes.
        TypeError
            If the inputs are not both DataFrames or both xarray Datasets.
        """

        if isinstance(data1, pd.DataFrame) and isinstance(data2, pd.DataFrame):
            if data1.shape != data2.shape:
                raise ValueError("DataFrames must have the same shape")
            variables = data1.columns
        elif isinstance(data1, xr.Dataset) and isinstance(data2, xr.Dataset):
            if sorted(list(data1.dims)) != sorted(list(data2.dims)) or sorted(
                list(data1.data_vars)
            ) != sorted(list(data2.data_vars)):
                raise ValueError(
                    "Datasets must have the same dimensions, coordinates and variables"
                )
            variables = data1.data_vars
        else:
            raise TypeError(
                "Inputs must be either both DataFrames or both xarray Datasets"
            )

        metrics = {}
        for var in variables:
            if isinstance(data1, pd.DataFrame):
                y_true = data1[var]
                y_pred = data2[var]
            else:
                y_true = data1[var].values.reshape(-1)
                y_pred = data2[var].values.reshape(-1)

            metrics[var] = {
                "mean_squared_error": mean_squared_error(y_true, y_pred),
                "r2_score": r2_score(y_true, y_pred),
                "mean_absolute_error": mean_absolute_error(y_true, y_pred),
                "explained_variance_score": explained_variance_score(y_true, y_pred),
            }

        return pd.DataFrame(metrics).T

    @staticmethod
    def get_uv_components(x_deg: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        This method calculates the u and v components for the given directional data.

        Here, we assume that the directional data is in degrees,
            beign 0° the North direction,
            and increasing clockwise.

                   0° N
                    |
                    |
        270° W <---------> 90° E
                    |
                    |
                  90° S

        Parameters
        ----------
        x_deg : np.ndarray
            The directional data in degrees.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            The u and v components.
        """

        return get_uv_components(x_deg)

    @staticmethod
    def get_degrees_from_uv(xu: np.ndarray, xv: np.ndarray) -> np.ndarray:
        """
        This method calculates the degrees from the u and v components.

        Here, we assume u and v represent angles between 0 and 360 degrees,
            where 0° is the North direction,
            and increasing clockwise.

                     (u=0, v=1)
                         |
                         |
        (u=-1, v=0) <---------> (u=1, v=0)
                         |
                         |
                     (u=0, v=-1)

        Parameters
        ----------
        xu : np.ndarray
            The u component.
        xv : np.ndarray
            The v component.

        Returns
        -------
        np.ndarray
            The degrees.
        """

        return get_degrees_from_uv(xu, xv)

    def set_omp_num_threads(self, num_threads: int) -> None:
        """
        Sets the number of threads for OpenMP.

        Parameters
        ----------
        num_threads : int
            The number of threads.

        Warning
        -----
        - This methos is under development.
        """

        os.environ["OMP_NUM_THREADS"] = str(num_threads)

        # Re-import numpy if it is already imported
        if "numpy" in sys.modules:
            importlib.reload(np)

    def get_num_processors_available(self) -> int:
        """
        Gets the number of processors available.

        Returns
        -------
        int
            The number of processors available.

        TODO
        ----
        - Check whether available processors are used or not.
        """

        return int(os.cpu_count() * 0.9)

    def set_num_processors_to_use(self, num_processors: int) -> None:
        """
        Sets the number of processors to use for parallel processing.

        Parameters
        ----------
        num_processors : int
            The number of processors to use.
            If -1, all available processors will be used.
        """

        # Retrieve the number of processors available
        num_processors_available = self.get_num_processors_available()

        # Check if the number of processors requested is valid
        if num_processors == -1:
            num_processors = num_processors_available
        elif num_processors <= 0:
            raise ValueError("Number of processors must be greater than 0")
        elif (num_processors - num_processors_available) < 2:
            raise ValueError(
                "Number of processors requested is less than 2 processors available"
            )

        # Set the number of processors to use
        self.num_workers = num_processors

    def parallel_execute(
        self,
        func: Callable,
        items: List[Any],
        num_workers: int,
        cpu_intensive: bool = False,
        **kwargs,
    ) -> List[T]:
        """
        Execute a function in parallel using concurrent.futures.

        Parameters
        ----------
        func : Callable
            Function to execute for each item.
        items : List[Any]
            List of items to process.
        num_workers : int
            Number of parallel workers.
        cpu_intensive : bool, optional
            Whether the function is CPU intensive. Default is False.
        **kwargs : dict
            Additional keyword arguments for func.

        Returns
        -------
        List[T]
            List of results from each function call

        Warnings
        --------
        - cpu_intensive = True does not work with non-pickable objects (Under development).
        """

        results = {}

        executor_class = ProcessPoolExecutor if cpu_intensive else ThreadPoolExecutor
        self.logger.info(f"Using {executor_class.__name__} for parallel execution")

        with executor_class(max_workers=num_workers) as executor:
            future_to_item = {
                executor.submit(func, *item, **kwargs)
                if isinstance(item, tuple)
                else executor.submit(func, item, **kwargs): i
                for i, item in enumerate(items)
            }
            for future in as_completed(future_to_item):
                i = future_to_item[future]
                try:
                    result = future.result()
                    results[i] = result
                except Exception as exc:
                    self.logger.error(f"Job for {i} generated an exception: {exc}")

        return results



import os

import numpy as np


def write_array_in_file(array: np.ndarray, filename: str) -> None:
    """
    Write an array in a file.

    Parameters
    ----------
    array : np.ndarray
        The array to be written. Can be 1D or 2D.
    filename : str
        The name of the file.
    """

    with open(filename, "w") as f:
        if array.ndim == 1:
            for item in array:
                f.write(f"{item}\n")
        elif array.ndim == 2:
            for row in array:
                f.write(" ".join(map(str, row)) + "\n")
        else:
            raise ValueError("Only 1D and 2D arrays are supported")


def copy_files(src: str, dst: str) -> None:
    """
    Copy file(s) from source to destination.

    Parameters
    ----------
    src : str
        The source file.
    dst : str
        The destination file.
    """

    if os.path.isdir(src):
        os.makedirs(dst, exist_ok=True)
        for file in os.listdir(src):
            with open(file, "r") as f:
                content = f.read()
            with open(os.path.join(dst, file), "w") as f:
                f.write(content)
    else:
        with open(src, "r") as f:
            content = f.read()
        with open(dst, "w") as f:
            f.write(content)



import os
import re
from typing import List, Tuple

import numpy as np
import pandas as pd
import xarray as xr
from scipy.signal import find_peaks

from ...waves.spectra import spectral_analysis
from ...waves.statistics import upcrossing
from .._base_wrappers import BaseModelWrapper


class SwashModelWrapper(BaseModelWrapper):
    """
    Wrapper for the SWASH model.
    https://swash.sourceforge.io/online_doc/swashuse/swashuse.html#input-and-output-files

    Attributes
    ----------
    default_parameters : dict
        The default parameters type for the wrapper.
    available_launchers : dict
        The available launchers for the wrapper.
    postprocess_functions : dict
        The postprocess functions for the wrapper.

    Methods
    -------
    list_available_postprocess_vars -> List[str]
        List available postprocess variables.
    _read_tabfile -> pd.DataFrame
        Read a tab file and return a pandas DataFrame.
    _convert_case_output_files_to_nc -> xr.Dataset
        Convert tab files to netCDF file.
    get_case_percentage_from_file -> float
        Get the case percentage from the output log file.
    monitor_cases -> pd.DataFrame
        Monitor the cases and log relevant information.
    postprocess_case -> xr.Dataset
        Convert tab ouput files to netCDF file.
    join_postprocessed_files -> xr.Dataset
        Join postprocessed files in a single Dataset.
    find_maximas -> Tuple[np.ndarray, np.ndarray]
        Find the individual maxima of an array.
    get_waterlevel -> xr.Dataset
        Get water level from the output netCDF file.
    calculate_runup2 -> xr.Dataset
        Calculates runup 2% (Ru2) from the output netCDF file.
    calculate_runup -> xr.Dataset
        Stores runup from the output netCDF file.
    calculate_setup -> xr.Dataset
        Calculates mean setup (Msetup) from the output netCDF file.
    calculate_statistical_analysis -> xr.Dataset
        Calculates zero-upcrossing analysis to obtain individual wave heights (Hi) and wave periods (Ti).
    calculate_spectral_analysis -> xr.Dataset
        Makes a water level spectral analysis (scipy.signal.welch)
        then separates incident waves, infragravity waves, very low frequency waves.
    """

    default_parameters = {
        "vegetation_height": float,
    }

    available_launchers = {
        "serial": "swash_serial.exe",
        "mpi": "mpirun -np 2 swash_mpi.exe",
        "docker_serial": "docker run --rm -v .:/case_dir -w /case_dir geoocean/rocky8 swash_serial.exe",
        "docker_mpi": "docker run --rm -v .:/case_dir -w /case_dir geoocean/rocky8 mpirun -np 2 swash_mpi.exe",
        "geoocean-cluster": "launchSwash.sh",
    }

    postprocess_functions = {
        "Ru2": "calculate_runup2",
        "RuDist": "calculate_runup",
        "Msetup": "calculate_setup",
        "Hrms": "calculate_statistical_analysis",
        "Hfreqs": "calculate_spectral_analysis",
        "Watlev": "get_waterlevel",
    }

    def __init__(
        self,
        templates_dir: str,
        model_parameters: dict,
        output_dir: str,
        templates_name: dict = "all",
        debug: bool = True,
    ) -> None:
        """
        Initialize the SWASH model wrapper.
        """

        super().__init__(
            templates_dir=templates_dir,
            model_parameters=model_parameters,
            output_dir=output_dir,
            templates_name=templates_name,
            default_parameters=self.default_parameters,
        )
        self.set_logger_name(
            name=self.__class__.__name__, level="DEBUG" if debug else "INFO"
        )

    def list_available_postprocess_vars(self) -> List[str]:
        """
        List available postprocess variables.

        Returns
        -------
        List[str]
            The available postprocess variables.
        """

        return list(self.postprocess_functions.keys())

    @staticmethod
    def _read_tabfile(file_path: str) -> pd.DataFrame:
        """
        Read a tab file and return a pandas DataFrame.
        This function is used to read the output files of SWASH.

        Parameters
        ----------
        file_path : str
            The file path.

        Returns
        -------
        pd.DataFrame
            The pandas DataFrame.
        """

        f = open(file_path, "r")
        lines = f.readlines()
        # read head colums (variables names)
        names = lines[4].split()
        names = names[1:]  # Eliminate '%'
        # read data rows
        values = pd.Series(lines[7:]).str.split(expand=True).values.astype(float)
        df = pd.DataFrame(values, columns=names)
        f.close()

        return df

    def _convert_case_output_files_to_nc(
        self, case_num: int, output_path: str, run_path: str
    ) -> xr.Dataset:
        """
        Convert tab files to netCDF file.

        Parameters
        ----------
        case_num : int
            The case number.
        output_path : str
            The output path.
        run_path : str
            The run path.

        Returns
        -------
        xr.Dataset
            The xarray Dataset.
        """

        df_output = self._read_tabfile(file_path=output_path)
        df_output.set_index(
            ["Xp", "Yp", "Tsec"], inplace=True
        )  # set index to Xp, Yp and Tsec
        ds_ouput = df_output.to_xarray()

        df_run = self._read_tabfile(file_path=run_path)
        df_run.set_index(["Tsec"], inplace=True)
        ds_run = df_run.to_xarray()

        # merge output files to one xarray.Dataset
        ds = xr.merge([ds_ouput, ds_run], compat="no_conflicts")

        # assign correct coordinate case_num
        ds.coords["case_num"] = case_num

        return ds

    def get_case_percentage_from_file(self, output_log_file: str) -> str:
        """
        Get the case percentage from the output log file.

        Parameters
        ----------
        output_log_file : str
            The output log file.

        Returns
        -------
        str
            The case percentage.
        """

        if not os.path.exists(output_log_file):
            return "0 %"

        progress_pattern = r"\[\s*(\d+)%\]"
        with open(output_log_file, "r") as f:
            for line in reversed(f.readlines()):
                match = re.search(progress_pattern, line)
                if match:
                    return f"{match.group(1)} %"

        return "0 %"  # if no progress is found

    def monitor_cases(self) -> pd.DataFrame:
        """
        Monitor the cases and log relevant information.

        Returns
        -------
        pd.DataFrame
            The cases percentage.
        """

        cases_percentage = {}

        for case_dir in self.cases_dirs:
            output_log_file = os.path.join(case_dir, "wrapper_out.log")
            progress = self.get_case_percentage_from_file(
                output_log_file=output_log_file
            )
            cases_percentage[os.path.basename(case_dir)] = progress

        return pd.DataFrame(cases_percentage.items(), columns=["Case", "Percentage"])

    def postprocess_case(
        self, case_num: int, case_dir: str, output_vars: List[str] = None
    ) -> xr.Dataset:
        """
        Convert tab output files to netCDF file.

        Parameters
        ----------
        case_num : int
            The case number.
        case_dir : str
            The case directory.
        output_vars : list, optional
            The output variables to postprocess. Default is None.

        Returns
        -------
        xr.Dataset
            The postprocessed Dataset.
        """

        import warnings

        warnings.filterwarnings("ignore")

        self.logger.info(f"Postprocessing case {case_num} in {case_dir}.")

        if output_vars is None:
            self.logger.info("Postprocessing all available variables.")
            output_vars = list(self.postprocess_functions.keys())

        output_nc_path = os.path.join(case_dir, "output.nc")
        if not os.path.exists(output_nc_path):
            # Convert tab files to netCDF file
            output_path = os.path.join(case_dir, "output.tab")
            run_path = os.path.join(case_dir, "run.tab")
            output_nc = self._convert_case_output_files_to_nc(
                case_num=case_num, output_path=output_path, run_path=run_path
            )
            output_nc.to_netcdf(output_nc_path)
        else:
            self.logger.info("Reading existing output.nc file.")
            output_nc = xr.open_dataset(output_nc_path)

        # Postprocess variables from output.nc
        var_ds_list = []
        for var in output_vars:
            if var in self.postprocess_functions:
                var_ds = getattr(self, self.postprocess_functions[var])(
                    case_num=case_num, case_dir=case_dir, output_nc=output_nc
                )
                var_ds_list.append(var_ds)
            else:
                self.logger.warning(
                    f"Variable {var} is not available for postprocessing."
                )

        # Merge all variables in one Dataset
        ds = xr.merge(var_ds_list, compat="no_conflicts")

        # Save Dataset to netCDF file
        ds.to_netcdf(os.path.join(case_dir, "output_postprocessed.nc"))

        return ds

    def join_postprocessed_files(
        self, postprocessed_files: List[xr.Dataset]
    ) -> xr.Dataset:
        """
        Join postprocessed files in a single Dataset.

        Parameters
        ----------
        postprocessed_files : list
            The postprocessed files.

        Returns
        -------
        xr.Dataset
            The joined Dataset.
        """

        return xr.concat(postprocessed_files, dim="case_num")

    def find_maximas(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Find the individual maxima of an array.

        Parameters
        ----------
        x : np.ndarray
            The array (should be the water level time series).

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            The peaks and the values of the peaks.
        """

        peaks, _ = find_peaks(x=x)

        return peaks, x[peaks]

    def get_waterlevel(
        self, case_num: int, case_dir: str, output_nc: xr.Dataset
    ) -> xr.Dataset:
        """
        Get water level from the output netCDF file.

        Parameters
        ----------
        case_num : int
            The case number.
        case_dir : str
            The case directory.
        output_nc : xr.Dataset
            The output netCDF file.

        Returns
        -------
        xr.Dataset
            The water level.
        """

        # get water level
        return output_nc[["Watlev"]].squeeze()

    def calculate_runup2(
        self, case_num: int, case_dir: str, output_nc: xr.Dataset
    ) -> xr.Dataset:
        """
        Calculates runup 2% (Ru2) from the output netCDF file.

        Parameters
        ----------
        case_num : int
            The case number.
        case_dir : str
            The case directory.
        output_nc : xr.Dataset
            The output netCDF file.

        Returns
        -------
        xr.Dataset
            The runup 2% (Ru2).
        """

        # get runup
        runup = output_nc["Runlev"].values

        # find individual wave uprushes
        _, val_peaks = self.find_maximas(runup)

        # calculate ru2
        ru2 = np.percentile(val_peaks, 98)

        # create xarray Dataset with ru2 value depending on case_num
        ds = xr.Dataset({"Ru2": ("case_num", [ru2])}, {"case_num": [case_num]})

        return ds

    def calculate_runup(
        self, case_num: int, case_dir: str, output_nc: xr.Dataset
    ) -> xr.Dataset:
        """
        Stores runup from the output netCDF file.

        Parameters
        ----------
        case_num : int
            The case number.
        case_dir : str
            The case directory.
        output_nc : xr.Dataset
            The output netCDF file.

        Returns
        -------
        xr.Dataset
            The runup.
        """

        # get runup
        ds = output_nc["Runlev"]

        return ds

    def calculate_setup(
        self, case_num: int, case_dir: str, output_nc: xr.Dataset
    ) -> xr.Dataset:
        """
        Calculates mean setup (Msetup) from the output netCDF file.

        Parameters
        ----------
        case_num : int
            The case number.
        case_dir : str
            The case directory.
        output_nc : xr.Dataset
            The output netCDF file.

        Returns
        -------
        xr.Dataset
            The mean setup (Msetup).
        """

        # create xarray Dataset with mean setup
        ds = output_nc["Watlev"].mean(dim="Tsec")
        ds = ds.to_dataset()

        # eliminate Yp dimension
        ds = ds.squeeze()

        # rename variable
        ds = ds.rename({"Watlev": "Msetup"})

        return ds

    def calculate_statistical_analysis(
        self, case_num: int, case_dir: str, output_nc: xr.Dataset
    ) -> xr.Dataset:
        """
        Calculates zero-upcrossing analysis to obtain individual wave heights (Hi) and wave periods (Ti).

        Parameters
        ----------
        case_num : int
            The case number.
        case_dir : str
            The case directory.
        output_nc : xr.Dataset
            The output netCDF file.

        Returns
        -------
        xr.Dataset
            The statistical analysis.
        """

        # for every X coordinate in domain
        df_Hrms = pd.DataFrame()

        for x in output_nc["Xp"].values:
            dsw = output_nc.sel(Xp=x)

            # obtain series of water level
            series_water = dsw["Watlev"].values
            time_series = dsw["Tsec"].values

            # perform statistical analysis
            # _, Hi = upcrossing(time_series, series_water)
            _, Hi = upcrossing(np.vstack([time_series, series_water]).T)

            # calculate Hrms
            Hrms_x = np.sqrt(np.mean(Hi**2))
            df_Hrms.loc[x, "Hrms"] = Hrms_x

        # convert pd DataFrame to xr Dataset
        df_Hrms.index.name = "Xp"
        ds = df_Hrms.to_xarray()

        # assign coordinate case_num
        ds = ds.assign_coords({"case_num": [output_nc["case_num"].values]})

        return ds

    def calculate_spectral_analysis(
        self, case_num: int, case_dir: str, output_nc: xr.Dataset
    ) -> xr.Dataset:
        """
        Makes a water level spectral analysis (scipy.signal.welch)
        then separates incident waves, infragravity waves, very low frequency waves.

        Parameters
        ----------
        case_num : int
            The case number.
        case_dir : str
            The case directory.
        output_nc : xr.Dataset
            The output netCDF file.

        Returns
        -------
        xr.Dataset
            The spectral analysis.
        """

        delttbl = np.diff(output_nc["Tsec"].values)[1]

        df_H_spectral = pd.DataFrame()

        for x in output_nc["Xp"].values:
            dsw = output_nc.sel(Xp=x)
            series_water = dsw["Watlev"].values

            # calculate significant, SS, IG and VLF wave heighs
            Hs, Hss, Hig, Hvlf = spectral_analysis(series_water, delttbl)

            df_H_spectral.loc[x, "Hs"] = Hs
            df_H_spectral.loc[x, "Hss"] = Hss
            df_H_spectral.loc[x, "ig"] = Hig
            df_H_spectral.loc[x, "Hvlf"] = Hvlf

        # convert pd DataFrame to xr Dataset
        df_H_spectral.index.name = "Xp"
        ds = df_H_spectral.to_xarray()

        # assign coordinate case_num
        ds = ds.assign_coords({"case_num": [output_nc["case_num"].values]})

        return ds



import numpy as np
from scipy import signal


def spectral_analysis(water_level, delta_time):
    """
    Performs spectral analisys of water level variable

    water_level - water level series (m)
    delta_time - time delta for sample (s)
    """

    # estimate power specrtal density
    f, E = signal.welch(
        water_level,
        fs=1 / delta_time,
        nfft=512,
        scaling="density",
    )
    E = np.reshape(E, -1)

    # Slice frequency to divide energy into components:

    # Incident waves IC
    fIC = f[np.where(f > 0.04)]
    EIC = E[np.where(f > 0.04)]
    mIC = np.trapz(EIC, x=fIC)

    # Infragravity waves IG
    fIG = f[(np.where(f > 0.004) and np.where(f < 0.04))]
    EIG = E[(np.where(f > 0.004) and np.where(f < 0.04))]
    mIG = np.trapz(EIG, x=fIG)

    # Very low frequency waves VLF
    fVLF = f[(np.where(f > 0.001) and np.where(f < 0.004))]
    EVLF = E[(np.where(f > 0.001) and np.where(f < 0.004))]
    mVLF = np.trapz(EVLF, x=fVLF)

    m0 = np.trapz(E, x=f)

    Hsi = 4 * np.sqrt(m0)
    HIC = 4 * np.sqrt(mIC)
    HIG = 4 * np.sqrt(mIG)
    HVLF = 4 * np.sqrt(mVLF)

    return Hsi, HIC, HIG, HVLF



import os.path as op

import numpy as np
import wavespectra
import xarray as xr
from wavespectra.construct import construct_partition

from bluemath_tk.waves.binwaves import (
    generate_swan_cases,
    process_kp_coefficients,
    reconstruc_spectra,
)
from bluemath_tk.wrappers.swan.swan_wrapper import SwanModelWrapper

example_directions = [
    1.5,
    4.5,
    7.5,
    10.5,
    13.5,
    16.5,
    19.5,
    22.5,
    25.5,
    28.5,
    31.5,
    34.5,
    37.5,
    40.5,
    43.5,
    46.5,
    49.5,
    52.5,
    55.5,
    58.5,
    61.5,
    64.5,
    67.5,
    70.5,
    73.5,
    76.5,
    79.5,
    82.5,
    85.5,
    88.5,
    91.5,
    94.5,
    97.5,
    100.5,
    103.5,
    106.5,
    109.5,
    112.5,
    115.5,
    118.5,
    121.5,
    124.5,
    127.5,
    130.5,
    133.5,
    136.5,
    139.5,
    142.5,
    145.5,
    148.5,
    151.5,
    154.5,
    157.5,
    160.5,
    163.5,
    166.5,
    169.5,
    172.5,
    175.5,
    178.5,
    181.5,
    184.5,
    187.5,
    190.5,
    193.5,
    196.5,
    199.5,
    202.5,
    205.5,
    208.5,
    211.5,
    214.5,
    217.5,
    220.5,
    223.5,
    226.5,
    229.5,
    232.5,
    235.5,
    238.5,
    241.5,
    244.5,
    247.5,
    250.5,
    253.5,
    256.5,
    259.5,
    262.5,
    265.5,
    268.5,
    271.5,
    274.5,
    277.5,
    280.5,
    283.5,
    286.5,
    289.5,
    292.5,
    295.5,
    298.5,
    301.5,
    304.5,
    307.5,
    310.5,
    313.5,
    316.5,
    319.5,
    322.5,
    325.5,
    328.5,
    331.5,
    334.5,
    337.5,
    340.5,
    343.5,
    346.5,
    349.5,
    352.5,
    355.5,
    358.5,
]
example_frequencies = [
    0.03,
    0.033,
    0.0363,
    0.0399,
    0.0438,
    0.0482,
    0.053,
    0.0582,
    0.064,
    0.0704,
    0.0774,
    0.0851,
    0.0935,
    0.1028,
    0.1131,
    0.1243,
    0.1367,
    0.1503,
    0.1652,
    0.1816,
    0.1997,
    0.2195,
    0.2413,
    0.2653,
    0.2917,
    0.3207,
    0.3526,
    0.3876,
    0.4262,
    0.4685,
    0.5151,
    0.5663,
    0.6226,
    0.6845,
    0.7525,
    0.8273,
    0.9096,
    1.0,
]

laura_directions = [
    262.5,
    247.5,
    232.5,
    217.5,
    202.5,
    187.5,
    172.5,
    157.5,
    142.5,
    127.5,
    112.5,
    97.5,
    82.5,
    67.5,
    52.5,
    37.5,
    22.5,
    7.5,
    352.5,
    337.5,
    322.5,
    307.5,
    292.5,
    277.5,
]
laura_frequencies = [
    0.03500004,
    0.03850004,
    0.04234991,
    0.04658508,
    0.05124342,
    0.05636788,
    0.06200474,
    0.06820491,
    0.07502551,
    0.082528,
    0.09078117,
    0.0998592,
    0.10984545,
    0.12082986,
    0.13291333,
    0.14620311,
    0.16082342,
    0.17690661,
    0.19459796,
    0.21405484,
    0.23546032,
    0.25900697,
    0.2849084,
    0.31340103,
    0.34474437,
    0.37921881,
    0.41713594,
    0.45884188,
    0.5047446,
]


def transform_CAWCR_WS(
    cawcr_dataset: xr.Dataset,
    subset_parameters: dict,
    available_case_num: np.ndarray,
) -> xr.Dataset:
    """
    Transform the wave spectra from CAWCR format to binwaves format.

    Parameters
    ----------
    cawcr_dataset : xr.Dataset
        The wave spectra dataset in CAWCR format.
    subset_parameters : dict
        A dictionary containing parameters for the subset processing.
    available_case_num : np.ndarray
        The available case numbers.

    Returns
    -------
    xr.Dataset
        The wave spectra dataset in binwaves format.
    """

    # First, reproject the wave spectra to the binwaves format
    ds = cawcr_dataset.rename({"frequency": "freq", "direction": "dir"})
    ds["efth"] = ds["efth"] * np.pi / 180.0
    ds["dir"] = ds["dir"] - 180.0
    ds["dir"] = np.where(ds["dir"] < 0, ds["dir"] + 360, ds["dir"])
    ds = ds.sortby("dir").sortby("freq")

    # Second, reproject into the available case numbers dimension
    case_num_spectra = []
    for case_num, case_dir, case_freq in zip(
        available_case_num,
        np.array(subset_parameters.get("dir"))[available_case_num],
        np.array(subset_parameters.get("freq"))[available_case_num],
    ):
        case_num_spectra.append(
            ds.efth.sel(freq=case_freq, dir=case_dir, method="nearest").expand_dims(
                {"case_num": [case_num]}
            )
        )
    ds_case_num = (
        xr.concat(case_num_spectra, dim="case_num").drop_vars("dir").drop_vars("freq")
    )

    return ds, ds_case_num


class BinWavesWrapper(SwanModelWrapper):
    """
    Wrapper example for the BinWaves model.
    """

    def build_case(self, case_dir: str, case_context: dict):
        input_spectrum = construct_partition(
            freq_name="jonswap",
            freq_kwargs={
                "freq": laura_frequencies,
                "fp": 1.0 / case_context.get("tp"),
                "hs": case_context.get("hs"),
            },
            dir_name="cartwright",
            dir_kwargs={
                "dir": laura_directions,
                "dm": case_context.get("dir"),
                "dspr": case_context.get("spr"),
            },
        )
        argmax_bin = np.argmax(input_spectrum.values)
        mono_spec_array = np.zeros(input_spectrum.freq.size * input_spectrum.dir.size)
        mono_spec_array[argmax_bin] = input_spectrum.sum(dim=["freq", "dir"])
        mono_spec_array = mono_spec_array.reshape(
            input_spectrum.freq.size, input_spectrum.dir.size
        )
        mono_input_spectrum = xr.Dataset(
            {
                "efth": (["freq", "dir"], mono_spec_array),
            },
            coords={
                "freq": input_spectrum.freq,
                "dir": input_spectrum.dir,
            },
        )
        wavespectra.SpecDataset(mono_input_spectrum).to_swan(
            op.join(case_dir, "input_spectra.bnd")
        )

    def build_cases(self, mode="one_by_one"):
        super().build_cases(mode)
        for case_dir, case_context in zip(self.cases_dirs, self.cases_context):
            self.build_case(case_dir, case_context)


# Usage example
if __name__ == "__main__":
    # Define the input templates and output directory
    templates_dir = (
        "/home/tausiaj/GitHub-GeoOcean/BlueMath/bluemath_tk/wrappers/swan/templates/"
    )
    templates_name = ["input.swn", "depth_main_cantabria.dat", "buoys.loc"]
    output_dir = "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_cases/swan/CAN_part/"
    # Generate swan model parameters
    model_parameters = (
        generate_swan_cases(
            directions_array=laura_directions,
            frequencies_array=laura_frequencies,
        )
        .astype(float)
        .to_dataframe()
        .reset_index()
        .to_dict(orient="list")
    )
    # Create an instance of the SWAN model wrapper
    swan_wrapper = BinWavesWrapper(
        templates_dir=templates_dir,
        templates_name=templates_name,
        model_parameters=model_parameters,
        output_dir=output_dir,
    )
    # Build the input files
    # swan_wrapper.build_cases(mode="one_by_one")
    # Set the cases directories from the output directory
    swan_wrapper.set_cases_dirs_from_output_dir()
    # List available launchers
    # print(swan_wrapper.list_available_launchers())
    # Run the model
    # swan_wrapper.run_cases(launcher="docker", parallel=True)
    # Post-process the output files
    # postprocessed_ds = swan_wrapper.postprocess_cases()
    # postprocessed_ds.to_netcdf(op.join(swan_wrapper.output_dir, "waves_part.nc"))
    # print(postprocessed_ds)
    # Get input and ouput spectra files from self.cases_dirs
    input_files = [op.join(d, "input_spectra.bnd") for d in swan_wrapper.cases_dirs]
    output_files = [op.join(d, "output.spec") for d in swan_wrapper.cases_dirs]
    # Extract binwaves kp coeffs
    kp_coeffs = process_kp_coefficients(
        list_of_input_spectra=input_files,
        list_of_output_spectra=output_files,
    )
    # Load interest spectra
    _, offshore_spectra = transform_CAWCR_WS(
        cawcr_dataset=xr.open_dataset(
            "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_data/ERA5_full.nc"
        ),
        subset_parameters=model_parameters,
        available_case_num=kp_coeffs.case_num.values,
    )
    # Reconstruct spectra
    onshore_spectra = reconstruc_spectra(
        offshore_spectra=offshore_spectra,
        kp_coeffs=kp_coeffs,
    )
    onshore_spectra.to_netcdf(op.join(swan_wrapper.output_dir, "onshore_spectra.nc"))
    print(onshore_spectra)



import numpy as np


def upcrossing(series):
    """
    Performs fast fourier transformation FFt by extracting
    the wave series mean and interpolating to find the cross-zero

    time   - time series
    watlev - surface elevation series

    return - Wave heigths and periods
    """

    # remove mean
    series[:, 1] -= np.mean(series[:, 1])

    # find 0's and calculate upcrossings
    zeros = np.where(np.diff(np.sign(series[:, 1])) > 0)[0]

    # calculate amplitudes (Hi) and periods (Ti)
    Ti = np.diff(zeros)
    Hi = [
        np.max(series[:, 1][zeros[i] : zeros[i + 1]])
        - np.min(series[:, 1][zeros[i] : zeros[i + 1]])
        for i in range(len(zeros) - 1)
    ]

    return Hi, Ti



from typing import List, Tuple

import numpy as np
import pandas as pd
from sklearn.cluster import KMeans

from ..core.decorators import validate_data_kma
from ._base_datamining import BaseClustering


class KMAError(Exception):
    """
    Custom exception for KMA class.
    """

    def __init__(self, message: str = "KMA error occurred."):
        self.message = message
        super().__init__(self.message)


class KMA(BaseClustering):
    """
    K-Means (KMA) class.

    This class performs the K-Means algorithm on a given dataframe.

    Attributes
    ----------
    num_clusters : int
        The number of clusters to use in the K-Means algorithm.
    seed : int
        The random seed to use as initial datapoint.
    data : pd.DataFrame
        The input data.
    normalized_data : pd.DataFrame
        The normalized input data.
    data_to_fit : pd.DataFrame
        The data to fit the K-Means algorithm.
    data_variables : List[str]
        A list with all data variables.
    directional_variables : List[str]
        A list with directional variables.
    fitting_variables : List[str]
        A list with fitting variables.
    custom_scale_factor : dict
        A dictionary of custom scale factors.
    scale_factor : dict
        A dictionary of scale factors (after normalizing the data).
    centroids : pd.DataFrame
        The selected centroids.
    normalized_centroids : pd.DataFrame
        The selected normalized centroids.
    centroid_real_indices : np.array
        The real indices of the selected centroids.

    Methods
    -------
    fit -> None
        Fit the K-Means algorithm to the provided data.
    predict -> Tuple[np.ndarray, pd.DataFrame]
        Predict the nearest centroid for the provided data.
    fit_predict -> Tuple[np.ndarray, pd.DataFrame]
        Fit the K-Means algorithm to the provided data and predict the nearest
        centroid for each data point.

    Notes
    -----
    - The K-Means algorithm is used to cluster data points into k clusters.
    - The K-Means algorithm is sensitive to the initial centroids.
    - The K-Means algorithm is not suitable for large datasets.

    Examples
    --------
    .. jupyter-execute::

        import numpy as np
        import pandas as pd
        from bluemath_tk.datamining.kma import KMA

        data = pd.DataFrame(
            {
                'Hs': np.random.rand(1000) * 7,
                'Tp': np.random.rand(1000) * 20,
                'Dir': np.random.rand(1000) * 360
            }
        )
        kma = KMA(num_clusters=5)
        nearest_centroids_idxs, nearest_centroids_df = kma.fit_predict(
            data=data,
            directional_variables=['Dir'],
        )

        kma.plot_selected_centroids(plot_text=True)
    """

    def __init__(
        self,
        num_clusters: int,
        seed: int = None,
    ) -> None:
        """
        Initializes the KMA class.

        Parameters
        ----------
        num_clusters : int
            The number of clusters to use in the K-Means algorithm.
            Must be greater than 0.
        seed : int, optional
            The random seed to use as initial datapoint.
            Must be greater or equal to 0 and less than number of datapoints.
            Default is 0.

        Raises
        ------
        ValueError
            If num_clusters is not greater than 0.
            Or if seed is not greater or equal to 0.
        """

        super().__init__()
        self.set_logger_name(name=self.__class__.__name__)
        if num_clusters > 0:
            self.num_clusters = int(num_clusters)
        else:
            raise ValueError("Variable num_clusters must be > 0")
        if seed is None:
            self.seed = None
        elif seed >= 0:
            self.seed = int(seed)
        else:
            raise ValueError("Variable seed must be >= 0")
        # TODO: check random_state and n_init
        self._kma = KMeans(
            n_clusters=self.num_clusters,
            random_state=self.seed,
        )
        self.logger.info(
            f"KMA object created with {self.num_clusters} clusters and seed {self.seed}."
            "To customize kma, do self.kma = dict(n_clusters=..., random_state=..., etc)"
        )
        self._data: pd.DataFrame = pd.DataFrame()
        self._normalized_data: pd.DataFrame = pd.DataFrame()
        self._data_to_fit: pd.DataFrame = pd.DataFrame()
        self.data_variables: List[str] = []
        self.directional_variables: List[str] = []
        self.fitting_variables: List[str] = []
        self.custom_scale_factor: dict = {}
        self.scale_factor: dict = {}
        self.centroids: pd.DataFrame = pd.DataFrame()
        self.normalized_centroids: pd.DataFrame = pd.DataFrame()
        self.centroid_real_indices: np.array = np.array([])
        self.is_fitted: bool = False

        self._exclude_attributes = [
            "_data",
            "_normalized_data",
            "_data_to_fit",
        ]

    @property
    def kma(self) -> KMeans:
        return self._kma

    @kma.setter
    def kma(self, **kwargs) -> None:
        self._kma = KMeans(**kwargs)

    @property
    def data(self) -> pd.DataFrame:
        return self._data

    @property
    def normalized_data(self) -> pd.DataFrame:
        return self._normalized_data

    @property
    def data_to_fit(self) -> pd.DataFrame:
        return self._data_to_fit

    @validate_data_kma
    def fit(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
        min_number_of_points: int = None,
    ) -> None:
        """
        Fit the K-Means algorithm to the provided data.

        This method initializes centroids for the K-Means algorithm using the
        provided dataframe and custom scale factor.
        It normalizes the data, and returns the calculated centroids.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the KMA algorithm.
        directional_variables : List[str], optional
            A list of directional variables (will be transformed to u and v).
            Default is [].
        custom_scale_factor : dict, optional
            A dictionary specifying custom scale factors for normalization.
            Default is {}.
        min_number_of_points : int, optional
            The minimum number of points to consider a cluster.
            Default is None.
        """

        self._data = data.copy()
        self.directional_variables = directional_variables.copy()
        for directional_variable in self.directional_variables:
            u_comp, v_comp = self.get_uv_components(
                x_deg=self.data[directional_variable].values
            )
            self.data[f"{directional_variable}_u"] = u_comp
            self.data[f"{directional_variable}_v"] = v_comp
        self.data_variables = list(self.data.columns)
        self.custom_scale_factor = custom_scale_factor.copy()

        # Get just the data to be used in the fitting
        self._data_to_fit = self.data.copy()
        for directional_variable in self.directional_variables:
            self.data_to_fit.drop(columns=[directional_variable], inplace=True)
        self.fitting_variables = list(self.data_to_fit.columns)

        # Normalize data using custom min max scaler
        self._normalized_data, self.scale_factor = self.normalize(
            data=self.data_to_fit, custom_scale_factor=self.custom_scale_factor
        )

        # Fit K-Means algorithm
        if min_number_of_points is not None:
            stable_kma_child = False
            number_of_tries = 0
            while not stable_kma_child:
                kma_child = KMeans(
                    n_clusters=self.num_clusters,
                    random_state=self.seed,
                )
                predicted_labels = kma_child.fit_predict(self.normalized_data)
                _unique_labels, counts = np.unique(predicted_labels, return_counts=True)
                if np.all(counts >= min_number_of_points):
                    stable_kma_child = True
                number_of_tries += 1
                if number_of_tries > 10:
                    raise ValueError(
                        "Failed to find a stable K-Means configuration after 10 attempts."
                    )
            self.logger.info(
                f"Found a stable K-Means configuration after {number_of_tries} attempts."
            )
            self._kma = kma_child
        else:
            self._kma = self.kma.fit(self.normalized_data)

        # Calculate the centroids
        self.centroid_real_indices = self.kma.labels_.copy()
        self.normalized_centroids = pd.DataFrame(
            self.kma.cluster_centers_, columns=self.fitting_variables
        )
        self.centroids = self.denormalize(
            normalized_data=self.normalized_centroids, scale_factor=self.scale_factor
        )
        for directional_variable in self.directional_variables:
            self.centroids[directional_variable] = self.get_degrees_from_uv(
                xu=self.centroids[f"{directional_variable}_u"].values,
                xv=self.centroids[f"{directional_variable}_v"].values,
            )

        # Set the fitted flag to True
        self.is_fitted = True

    def predict(self, data: pd.DataFrame) -> Tuple[np.ndarray, pd.DataFrame]:
        """
        Predict the nearest centroid for the provided data.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the prediction.

        Returns
        -------
        Tuple[np.ndarray, pd.DataFrame]
            A tuple containing the nearest centroid index for each data point,
            and the nearest centroids.
        """

        if self.is_fitted is False:
            raise KMAError("KMA model is not fitted.")

        data = data.copy()  # Avoid modifying the original data to predict
        for directional_variable in self.directional_variables:
            u_comp, v_comp = self.get_uv_components(
                x_deg=data[directional_variable].values
            )
            data[f"{directional_variable}_u"] = u_comp
            data[f"{directional_variable}_v"] = v_comp
            data.drop(columns=[directional_variable], inplace=True)
        normalized_data, _ = self.normalize(
            data=data, custom_scale_factor=self.scale_factor
        )
        y = self.kma.predict(X=normalized_data)

        return y, self.centroids.iloc[y]

    def fit_predict(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
        min_number_of_points: int = None,
    ) -> Tuple[np.ndarray, pd.DataFrame]:
        """
        Fit the K-Means algorithm to the provided data and predict the nearest centroid
        for each data point.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the KMA algorithm.
        directional_variables : List[str], optional
            A list of directional variables (will be transformed to u and v).
            Default is [].
        custom_scale_factor : dict
            A dictionary specifying custom scale factors for normalization.
            Default is {}.
        min_number_of_points : int, optional
            The minimum number of points to consider a cluster.
            Default is None.

        Returns
        -------
        Tuple[np.ndarray, pd.DataFrame]
            A tuple containing the nearest centroid index for each data point,
            and the nearest centroids.
        """

        self.fit(
            data=data,
            directional_variables=directional_variables,
            custom_scale_factor=custom_scale_factor,
            min_number_of_points=min_number_of_points,
        )

        return self.predict(data=data)



import unittest

import numpy as np

from bluemath_tk.core.data.sample_data import get_2d_dataset
from bluemath_tk.datamining.pca import PCA


def replace_nans_with_previous(arr: np.ndarray, fill_value: float = 0.0) -> np.ndarray:
    """
    Replaces NaNs in a 2D NumPy array with the previous value in the same column.
    If the first value in a column is NaN, it is replaced with `fill_value`.

    Parameters
    ----------
    arr : np.ndarray
        Input 2D array with NaN values.
    fill_value : float, optional
        Value to replace leading NaNs. Default is 0.0.

    Returns
    -------
    np.ndarray
        Array with NaNs replaced.
    """

    arr = arr.copy()  # Avoid modifying the original array
    mask = np.isnan(arr)

    # Replace NaNs in the first row with a default fill_value
    arr[0, mask[0, :]] = fill_value

    # Replace NaNs with the previous row's value
    for i in range(1, arr.shape[0]):  # Iterate over rows
        arr[i, :] = np.where(mask[i, :], arr[i - 1, :], arr[i, :])

    return arr


class TestPCA(unittest.TestCase):
    def setUp(self):
        self.ds = get_2d_dataset()
        self.pca = PCA(n_components=5, debug=True)
        self.ipca = PCA(n_components=5, is_incremental=True)

    def test_fit_transform(self):
        pcs = self.pca.fit_transform(
            data=self.ds,
            vars_to_stack=["X", "Y"],
            coords_to_stack=["coord1", "coord2"],
            pca_dim_for_rows="coord3",
            windows_in_pca_dim_for_rows={"X": [3], "Y": [1]},
            value_to_replace_nans={
                "X": 0.0,
                "X_3": 1.0,
                "Y": replace_nans_with_previous,
            },
            nan_threshold_to_drop={"X": 0.5, "Y": 0.5},
        )
        self.assertEqual(self.pca.is_fitted, True)
        self.assertEqual(pcs.PCs.shape[1], 5)
        self.assertEqual(pcs.PCs.shape[0], self.ds.sizes["coord3"])
        self.assertCountEqual(self.pca.eofs.data_vars, ["X", "X_3", "Y", "Y_1"])

    def test_inverse_transform(self):
        pcs = self.pca.fit_transform(
            data=self.ds,
            vars_to_stack=["X", "Y"],
            coords_to_stack=["coord1", "coord2"],
            pca_dim_for_rows="coord3",
            scale_data=False,
        )
        reconstructed_ds = self.pca.inverse_transform(PCs=pcs.isel(coord3=slice(0, 5)))
        self.assertAlmostEqual(
            self.ds.isel(coord1=5, coord2=5, coord3=1),
            reconstructed_ds.isel(coord1=5, coord2=5, coord3=1),
        )

    def test_incremental_fit(self):
        self.ipca.fit(
            data=self.ds,
            vars_to_stack=["X", "Y"],
            coords_to_stack=["coord1", "coord2"],
            pca_dim_for_rows="coord3",
        )
        self.assertEqual(self.ipca.is_fitted, True)


if __name__ == "__main__":
    unittest.main()



from abc import abstractmethod
from typing import List, Tuple

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from ..core.models import BlueMathModel
from ..core.plotting.base_plotting import DefaultStaticPlotting


class BaseSampling(BlueMathModel):
    """
    Base class for all sampling BlueMath models.
    This class provides the basic structure for all sampling models.

    Methods
    -------
    generate : pd.DataFrame
        Generates samples.
    plot_generated_data : Tuple[plt.figure, plt.axes]
        Plots the generated data on a scatter plot matrix.
    """

    @abstractmethod
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def generate(self, *args, **kwargs) -> pd.DataFrame:
        """
        Generates samples.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        pd.DataFrame
            The generated samples.
        """

        return pd.DataFrame()

    def plot_generated_data(
        self,
        data_color: str = "blue",
        **kwargs,
    ) -> Tuple[plt.figure, plt.axes]:
        """
        Plots the generated data on a scatter plot matrix.

        Parameters
        ----------
        data_color : str, optional
            Color for the data points. Default is "blue".
        **kwargs : dict, optional
            Additional keyword arguments to be passed to the scatter plot function.

        Returns
        -------
        plt.figure
            The figure object containing the plot.
        plt.axes
            Array of axes objects for the subplots.

        Raises
        ------
        ValueError
            If the data is empty.
        """

        if not self.data.empty:
            variables_names = list(self.data.columns)
            num_variables = len(variables_names)
        else:
            raise ValueError("Data must be a non-empty DataFrame with columns to plot.")

        # Create figure and axes
        default_static_plot = DefaultStaticPlotting()
        fig, axes = default_static_plot.get_subplots(
            nrows=num_variables - 1,
            ncols=num_variables - 1,
            sharex=False,
            sharey=False,
        )

        for c1, v1 in enumerate(variables_names[1:]):
            for c2, v2 in enumerate(variables_names[:-1]):
                default_static_plot.plot_scatter(
                    ax=axes[c2, c1],
                    x=self.data[v1],
                    y=self.data[v2],
                    c=data_color,
                    **kwargs,
                )
                if c1 == c2:
                    axes[c2, c1].set_xlabel(variables_names[c1 + 1])
                    axes[c2, c1].set_ylabel(variables_names[c2])
                elif c1 > c2:
                    axes[c2, c1].xaxis.set_ticklabels([])
                    axes[c2, c1].yaxis.set_ticklabels([])
                else:
                    fig.delaxes(axes[c2, c1])

        return fig, axes


class BaseClustering(BlueMathModel):
    """
    Base class for all clustering BlueMath models.
    This class provides the basic structure for all clustering models.

    Methods
    -------
    fit : None
        Fits the model to the data.
    predict : pd.DataFrame
        Predicts the clusters for the provided data.
    fit_predict : pd.DataFrame
        Fits the model to the data and predicts the clusters.
    plot_selected_centroids : Tuple[plt.figure, plt.axes]
        Plots data and selected centroids on a scatter plot matrix.
    plot_data_as_clusters : Tuple[plt.figure, plt.axes]
        Plots data as nearest clusters.
    """

    @abstractmethod
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def fit(self, *args, **kwargs) -> None:
        """
        Fits the model to the data.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.
        """

        pass

    @abstractmethod
    def predict(self, *args, **kwargs) -> pd.DataFrame:
        """
        Predicts the clusters for the provided data.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        pd.DataFrame
            The predicted clusters.
        """

        return pd.DataFrame()

    @abstractmethod
    def fit_predict(self, *args, **kwargs) -> pd.DataFrame:
        """
        Fits the model to the data and predicts the clusters.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        pd.DataFrame
            The predicted clusters.
        """

        return pd.DataFrame()

    def plot_selected_centroids(
        self,
        data_color: str = "blue",
        centroids_color: str = "red",
        plot_text: bool = False,
        **kwargs,
    ) -> Tuple[plt.figure, plt.axes]:
        """
        Plots data and selected centroids on a scatter plot matrix.

        Parameters
        ----------
        data_color : str, optional
            Color for the data points. Default is "blue".
        centroids_color : str, optional
            Color for the centroid points. Default is "red".
        plot_text : bool, optional
            Whether to display text labels for centroids. Default is False.
        **kwargs : dict, optional
            Additional keyword arguments to be passed to the scatter plot function.

        Returns
        -------
        plt.figure
            The figure object containing the plot.
        plt.axes
            Array of axes objects for the subplots.

        Raises
        ------
        ValueError
            If the data and centroids do not have the same number of columns or if the columns are empty.
        """

        if (
            len(self.data.columns) == len(self.centroids.columns)
            and list(self.data.columns) != []
        ):
            variables_names = list(self.data.columns)
            num_variables = len(variables_names)
        else:
            raise ValueError(
                "Data and centroids must have the same number of columns > 0."
            )

        # Create figure and axes
        default_static_plot = DefaultStaticPlotting()
        fig, axes = default_static_plot.get_subplots(
            nrows=num_variables - 1,
            ncols=num_variables - 1,
            sharex=False,
            sharey=False,
        )

        for c1, v1 in enumerate(variables_names[1:]):
            for c2, v2 in enumerate(variables_names[:-1]):
                default_static_plot.plot_scatter(
                    ax=axes[c2, c1],
                    x=self.data[v1],
                    y=self.data[v2],
                    c=data_color,
                    alpha=0.6,
                    **kwargs,
                )
                if self.centroids is not None:
                    default_static_plot.plot_scatter(
                        ax=axes[c2, c1],
                        x=self.centroids[v1],
                        y=self.centroids[v2],
                        c=centroids_color,
                        alpha=0.9,
                        **kwargs,
                    )
                    if plot_text:
                        for i in range(self.centroids.shape[0]):
                            axes[c2, c1].text(
                                self.centroids[v1][i],
                                self.centroids[v2][i],
                                str(i + 1),
                                fontsize=12,
                                fontweight="bold",
                            )
                if c1 == c2:
                    axes[c2, c1].set_xlabel(variables_names[c1 + 1])
                    axes[c2, c1].set_ylabel(variables_names[c2])
                elif c1 > c2:
                    axes[c2, c1].xaxis.set_ticklabels([])
                    axes[c2, c1].yaxis.set_ticklabels([])
                else:
                    fig.delaxes(axes[c2, c1])

        return fig, axes

    def plot_data_as_clusters(
        self,
        data: pd.DataFrame,
        nearest_centroids: np.ndarray,
        **kwargs,
    ) -> Tuple[plt.figure, plt.axes]:
        """
        Plots data as nearest clusters.

        Parameters
        ----------
        data : pd.DataFrame
            The data to plot.
        nearest_centroids : np.ndarray
            The nearest centroids.
        **kwargs : dict, optional
            Additional keyword arguments to be passed to the scatter plot function.

        Returns
        -------
        plt.figure
            The figure object containing the plot.
        plt.axes
            The axes object for the plot.
        """

        if (
            not data.empty
            and list(self.data.columns) != []
            and nearest_centroids.size > 0
        ):
            variables_names = list(data.columns)
            num_variables = len(variables_names)
        else:
            raise ValueError(
                "Data must have columns and nearest centroids must have values."
            )

        # Create figure and axes
        default_static_plot = DefaultStaticPlotting()
        fig, axes = default_static_plot.get_subplots(
            nrows=num_variables - 1,
            ncols=num_variables - 1,
            sharex=False,
            sharey=False,
        )

        # Gets colors for clusters and append to each nearest centroid
        colors_for_clusters = default_static_plot.get_list_of_colors_for_colormap(
            cmap="jet", num_colors=self.centroids.shape[0]
        )
        nearest_centroids_colors = [colors_for_clusters[i] for i in nearest_centroids]

        for c1, v1 in enumerate(variables_names[1:]):
            for c2, v2 in enumerate(variables_names[:-1]):
                default_static_plot.plot_scatter(
                    ax=axes[c2, c1],
                    x=data[v1],
                    y=data[v2],
                    c=nearest_centroids_colors,
                    alpha=0.9,
                    **kwargs,
                )
                if c1 == c2:
                    axes[c2, c1].set_xlabel(variables_names[c1 + 1])
                    axes[c2, c1].set_ylabel(variables_names[c2])
                elif c1 > c2:
                    axes[c2, c1].xaxis.set_ticklabels([])
                    axes[c2, c1].yaxis.set_ticklabels([])
                else:
                    fig.delaxes(axes[c2, c1])

        return fig, axes


class BaseReduction(BlueMathModel):
    """
    Base class for all dimensionality reduction BlueMath models.
    This class provides the basic structure for all dimensionality reduction models.
    """

    @abstractmethod
    def __init__(self) -> None:
        super().__init__()


class ClusteringComparator:
    """
    Class for comparing clustering models.
    """

    def __init__(self, list_of_models: List[BaseClustering]) -> None:
        """
        Initializes the ClusteringComparator class.
        """

        self.list_of_models = list_of_models

    def fit(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
    ) -> None:
        """
        Fits the clustering models.
        """

        for model in self.list_of_models:
            if model.__class__.__name__ == "SOM":
                model.fit(
                    data=data,
                    directional_variables=directional_variables,
                )
            else:
                model.fit(
                    data=data,
                    directional_variables=directional_variables,
                    custom_scale_factor=custom_scale_factor,
                )

    def plot_selected_centroids(self) -> None:
        """
        Plots the selected centroids for the clustering models.
        """

        for model in self.list_of_models:
            fig, axes = model.plot_selected_centroids()
            fig.suptitle(f"Selected centroids for {model.__class__.__name__}")

    def plot_data_as_clusters(self, data: pd.DataFrame) -> None:
        """
        Plots the data as clusters for the clustering models.
        """

        for model in self.list_of_models:
            nearest_centroids, _ = model.predict(data=data)
            fig, axes = model.plot_data_as_clusters(
                data=data, nearest_centroids=nearest_centroids
            )
            fig.suptitle(f"Data as clusters for {model.__class__.__name__}")



import os

import psutil
from dask.distributed import Client, LocalCluster


def get_total_ram() -> int:
    """
    Get the total RAM in the system.

    Returns
    -------
    int
        The total RAM in bytes.
    """

    return psutil.virtual_memory().total


def get_available_ram() -> int:
    """
    Get the available RAM in the system.

    Returns
    -------
    int
        The available RAM in bytes.
    """

    return psutil.virtual_memory().available


def setup_dask_client(n_workers: int = None, memory_limit: str = 0.5):
    """
    Setup a Dask client with controlled resources.

    Parameters
    ----------
    n_workers : int, optional
        Number of workers. Default is None.
    memory_limit : str, optional
        Memory limit per worker. Default is 0.5.

    Returns
    -------
    Client
        Dask distributed client

    Notes
    -----
    - Resources might vary depending on the hardware and the load of the machine.
      Be very careful when setting the number of workers and memory limit, as it
      might affect the performance of the machine, or in the worse case scenario,
      the performance of other users in the same machine (cluster case).
    """

    if n_workers is None:
        n_workers = int(os.environ.get("BLUEMATH_NUM_WORKERS", "1"))
    if isinstance(memory_limit, float):
        memory_limit *= get_available_ram() / get_total_ram()

    cluster = LocalCluster(
        n_workers=n_workers, threads_per_worker=1, memory_limit=memory_limit
    )
    client = Client(cluster)

    return client



import logging
from typing import Tuple, Union

import numpy as np
import pandas as pd
import pyproj
import xarray as xr
from sklearn.preprocessing import StandardScaler

available_projections = {
    "SPAIN": pyproj.Proj(proj="utm", zone=30, ellps="WGS84"),
}


def normalize(
    data: Union[pd.DataFrame, xr.Dataset],
    custom_scale_factor: dict = {},
    logger: logging.Logger = None,
) -> Tuple[Union[pd.DataFrame, xr.Dataset], dict]:
    """
    Normalize data to 0-1 using min max scaler approach.

    Parameters
    ----------
    data : pd.DataFrame or xr.Dataset
        Input data to be normalized.
    custom_scale_factor : dict, optional
        Dictionary with variables as keys and a list with two values as
        values. The first value is the minimum and the second value is the
        maximum used to normalize the variable. If not provided, the
        minimum and maximum values of the variable are used.
    logger : logging.Logger, optional
        Logger object to log warnings if the custom min or max is bigger or
        lower than the datapoints.

    Returns
    -------
    normalized_data : pd.DataFrame or xr.Dataset
        Normalized data.
    scale_factor : dict
        Dictionary with variables as keys and a list with two values as
        values. The first value is the minimum and the second value is the
        maximum used to normalize the variable.

    Notes
    -----
    - This method does not modify the input data, it creates a copy of the
      dataframe / dataset and normalizes it.
    - The normalization is done variable by variable, i.e. the minimum and
      maximum values are calculated for each variable.
    - If custom min or max is bigger or lower than the datapoints, it will
      be changed to the minimum or maximum of the datapoints and a warning
      will be logged.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from bluemath_tk.core.operations import normalize
    >>> df = pd.DataFrame(
    ...     {
    ...         "Hs": np.random.rand(1000) * 7,
    ...         "Tp": np.random.rand(1000) * 20,
    ...         "Dir": np.random.rand(1000) * 360,
    ...     }
    ... )
    >>> normalized_data, scale_factor = normalize(data=df)
    >>> import numpy as np
    >>> import xarray as xr
    >>> from bluemath_tk.core.data import normalize
    >>> ds = xr.Dataset(
    ...     {
    ...         "Hs": (("time",), np.random.rand(1000) * 7),
    ...         "Tp": (("time",), np.random.rand(1000) * 20),
    ...         "Dir": (("time",), np.random.rand(1000) * 360),
    ...     },
    ...     coords={"time": pd.date_range("2000-01-01", periods=1000)},
    ... )
    >>> normalized_data, scale_factor = normalize(data=ds)
    """

    if isinstance(data, pd.DataFrame):
        vars_to_normalize = list(data.columns)
    elif isinstance(data, xr.Dataset):
        vars_to_normalize = list(data.data_vars)
    else:
        raise TypeError("Data must be a pandas DataFrame or an xarray Dataset")
    normalized_data = data.copy()  # Copy data to avoid bad memory replacements
    scale_factor = (
        custom_scale_factor.copy()
    )  # Copy dict to avoid bad memory replacements
    for data_var in vars_to_normalize:
        data_var_min = normalized_data[data_var].min()
        data_var_max = normalized_data[data_var].max()
        if custom_scale_factor.get(data_var):
            if custom_scale_factor.get(data_var)[0] > data_var_min:
                if logger is not None:
                    logger.warning(
                        f"Proposed min custom scaler for {data_var} is bigger than datapoint"  # , using smallest datapoint
                    )
                else:
                    print(
                        f"Proposed min custom scaler for {data_var} is bigger than datapoint"  # , using smallest datapoint
                    )
            #     scale_factor[data_var][0] = data_var_min
            # else:
            data_var_min = custom_scale_factor.get(data_var)[0]
            if custom_scale_factor.get(data_var)[1] < data_var_max:
                if logger is not None:
                    logger.warning(
                        f"Proposed max custom scaler for {data_var} is lower than datapoint"  # , using biggest datapoint
                    )
                else:
                    print(
                        f"Proposed max custom scaler for {data_var} is lower than datapoint"  # , using biggest datapoint
                    )
            #     scale_factor[data_var][1] = data_var_max
            # else:
            data_var_max = custom_scale_factor.get(data_var)[1]
        else:
            scale_factor[data_var] = [data_var_min, data_var_max]
        normalized_data[data_var] = (normalized_data[data_var] - data_var_min) / (
            data_var_max - data_var_min
        )
    return normalized_data, scale_factor


def denormalize(
    normalized_data: Union[pd.DataFrame, xr.Dataset],
    scale_factor: dict,
) -> Union[pd.DataFrame, xr.Dataset]:
    """
    Denormalize data using provided scale_factor.

    Parameters
    ----------
    normalized_data : pd.DataFrame or xr.Dataset
        Input data that has been normalized and needs to be denormalized.
    scale_factor : dict
        Dictionary with variables as keys and a list with two values as
        values. The first value is the minimum and the second value is the
        maximum used to denormalize the variable.

    Returns
    -------
    data : pd.DataFrame or xr.Dataset
        Denormalized data.

    Notes
    -----
    - This method does not modify the input data, it creates a copy of the
      dataframe / dataset and denormalizes it.
    - The denormalization is done variable by variable, i.e. the minimum and
      maximum values are used to scale the data back to its original range.
    - Assumes that the scale_factor dictionary contains appropriate min and
      max values for each variable in the normalized_data.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from bluemath_tk.core.operation import denormalize
    >>> df = pd.DataFrame(
    ...     {
    ...         "Hs": np.random.rand(1000),
    ...         "Tp": np.random.rand(1000),
    ...         "Dir": np.random.rand(1000),
    ...     }
    ... )
    >>> scale_factor = {
    ...     "Hs": [0, 7],
    ...     "Tp": [0, 20],
    ...     "Dir": [0, 360],
    ... }
    >>> denormalized_data = denormalize(normalized_data=df, scale_factor=scale_factor)
    >>> import numpy as np
    >>> import xarray as xr
    >>> from bluemath_tk.core.operations import denormalize
    >>> ds = xr.Dataset(
    ...     {
    ...         "Hs": (("time",), np.random.rand(1000)),
    ...         "Tp": (("time",), np.random.rand(1000)),
    ...         "Dir": (("time",), np.random.rand(1000)),
    ...     },
    ...     coords={"time": pd.date_range("2000-01-01", periods=1000)},
    ... )
    >>> scale_factor = {
    ...     "Hs": [0, 7],
    ...     "Tp": [0, 20],
    ...     "Dir": [0, 360],
    ... }
    >>> denormalized_data = denormalize(normalized_data=ds, scale_factor=scale_factor)
    """

    if isinstance(normalized_data, pd.DataFrame):
        vars_to_denormalize = list(normalized_data.columns)
    elif isinstance(normalized_data, xr.Dataset):
        vars_to_denormalize = list(normalized_data.data_vars)
    else:
        raise TypeError("Data must be a pandas DataFrame or an xarray Dataset")
    data = normalized_data.copy()  # Copy data to avoid bad memory replacements
    for data_var in vars_to_denormalize:
        data[data_var] = (
            data[data_var] * (scale_factor[data_var][1] - scale_factor[data_var][0])
            + scale_factor[data_var][0]
        )
    return data


def standarize(
    data: Union[np.ndarray, pd.DataFrame, xr.Dataset],
    scaler: StandardScaler = None,
    transform: bool = False,
) -> Tuple[Union[np.ndarray, pd.DataFrame, xr.Dataset], StandardScaler]:
    """
    Standarize data to have mean 0 and std 1.

    Parameters
    ----------
    data : np.ndarray, pd.DataFrame or xr.Dataset
        Input data to be standarized.
    scaler : StandardScaler, optional
        Scaler object to use for standarization. Default is None.
    transform : bool
        Whether to just transform the data. Default to False.

    Returns
    -------
    standarized_data : np.ndarray, pd.DataFrame or xr.Dataset
        Standarized data.
    scaler : StandardScaler
        Scaler object used for standarization.

    Examples
    --------
    >>> import numpy as np
    >>> from bluemath_tk.core.operations import standarize
    >>> data = np.random.rand(1000, 3) * 10.0
    >>> standarized_data, scaler = standarize(data=data)
    """

    scaler = scaler or StandardScaler()
    if isinstance(data, np.ndarray):
        if transform:
            standarized_data = scaler.transform(X=data)
        else:
            standarized_data = scaler.fit_transform(X=data)
    elif isinstance(data, pd.DataFrame):
        if transform:
            standarized_data = scaler.transform(X=data.values)
        else:
            standarized_data = scaler.fit_transform(X=data.values)
        standarized_data = pd.DataFrame(standarized_data, columns=data.columns)
    elif isinstance(data, xr.Dataset):
        if transform:
            standarized_data = scaler.transform(X=data.to_array().values)
        else:
            standarized_data = scaler.fit_transform(X=data.to_array().values)
        standarized_data = xr.Dataset(
            {
                var_name: (tuple(data.coords), standarized_data[i_var])
                for i_var, var_name in enumerate(data.data_vars)
            },
            coords=data.coords,
        )
    return standarized_data, scaler


def destandarize(
    standarized_data: Union[np.ndarray, pd.DataFrame, xr.Dataset],
    scaler: StandardScaler,
) -> Union[np.ndarray, pd.DataFrame, xr.Dataset]:
    """
    Destandarize data using provided scaler.

    Parameters
    ----------
    standarized_data : np.ndarray, pd.DataFrame or xr.Dataset
        Standarized data to be destandarized.
    scaler : StandardScaler
        Scaler object used for standarization.

    Returns
    -------
    np.ndarray, pd.DataFrame or xr.Dataset
        Destandarized data.

    Examples
    --------
    >>> import numpy as np
    >>> from bluemath_tk.core.data import standarize, destandarize
    >>> data = np.random.rand(1000, 3) * 10.0
    >>> standarized_data, scaler = standarize(data=data)
    >>> data = destandarize(standarized_data=standarized_data, scaler=scaler)
    """

    if isinstance(standarized_data, np.ndarray):
        data = scaler.inverse_transform(X=standarized_data)
    elif isinstance(standarized_data, pd.DataFrame):
        data = scaler.inverse_transform(X=standarized_data.values)
        data = pd.DataFrame(data, columns=standarized_data.columns)
    elif isinstance(standarized_data, xr.Dataset):
        data = scaler.inverse_transform(X=standarized_data.to_array().values)
        data = xr.Dataset(
            {
                var_name: (tuple(standarized_data.coords), data[i_var])
                for i_var, var_name in enumerate(standarized_data.data_vars)
            },
            coords=standarized_data.coords,
        )
    return data


def get_uv_components(x_deg: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    This method calculates the u and v components for the given directional data.

    Here, we assume that the directional data is in degrees,
        beign 0° the North direction,
        and increasing clockwise.

                0° N
                |
                |
    270° W <---------> 90° E
                |
                |
                90° S

    Parameters
    ----------
    x_deg : np.ndarray
        The directional data in degrees.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The u and v components.
    """

    # Convert degrees to radians and adjust by subtracting from π/2
    x_rad = x_deg * np.pi / 180

    # Calculate x and y components using cosine and sine
    xu = np.sin(x_rad)
    xv = np.cos(x_rad)

    # Return the u and v components
    return xu, xv


def get_degrees_from_uv(xu: np.ndarray, xv: np.ndarray) -> np.ndarray:
    """
    This method calculates the degrees from the u and v components.

    Here, we assume u and v represent angles between 0 and 360 degrees,
        where 0° is the North direction,
        and increasing clockwise.

                 (u=0, v=1)
                     |
                     |
    (u=-1, v=0) <---------> (u=1, v=0)
                     |
                     |
                 (u=0, v=-1)

    Parameters
    ----------
    xu : np.ndarray
        The u component.
    xv : np.ndarray
        The v component.

    Returns
    -------
    np.ndarray
        The degrees.
    """

    # Calculate the degrees using the arctangent function
    x_deg = np.arctan2(xu, xv) * 180 / np.pi % 360

    # Return the degrees
    return x_deg


def convert_utm_to_lonlat(
    utm_x: np.ndarray,
    utm_y: np.ndarray,
    projection: Union[int, str, dict, pyproj.CRS],
) -> Tuple[np.ndarray, np.ndarray]:
    """
    This method converts UTM coordinates to Longitude and Latitude.

    Parameters
    ----------
    utm_x : np.ndarray
        The x values in UTM.
    utm_y : np.ndarray
        The y values in UTM.
    projection : int, str, dict, pyproj.CRS
        The projection to use for the transformation.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The longitude and latitude values.
    """

    if isinstance(projection, str):
        projection = available_projections.get(projection, projection)

    # Transform the UTM to LonLat coordinates
    reshape = False
    if utm_x.size != utm_y.size:
        reshape_size = (utm_y.size, utm_x.size)
        utm_x, utm_y = (
            np.meshgrid(utm_x, utm_y)[0].reshape(-1),
            np.meshgrid(utm_x, utm_y)[1].reshape(-1),
        )
        reshape = True
    lon, lat = projection(utm_x, utm_y, inverse=True)
    if reshape:
        lon, lat = (
            lon.reshape(*reshape_size)[0, :],
            lat.reshape(*reshape_size)[:, 0],
        )

    # Return the LonLat coordinates
    return lon, lat


def convert_lonlat_to_utm(
    lon: np.ndarray,
    lat: np.ndarray,
    projection: Union[int, str, dict, pyproj.CRS],
) -> Tuple[np.ndarray, np.ndarray]:
    """
    This method converts Longitude and Latitude to UTM coordinates.

    Parameters
    ----------
    lon : np.ndarray
        The longitude values.
    lat : np.ndarray
        The latitude values.
    projection : int, str, dict, pyproj.CRS
        The projection to use for the transformation.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The x and y coordinates in UTM.
    """

    if isinstance(projection, str):
        projection = available_projections.get(projection, projection)

    # Transform the LonLat to UTM coordinates
    reshape = False
    if lon.size != lat.size:
        reshape_size = (lat.size, lon.size)
        lon, lat = (
            np.meshgrid(lon, lat)[0].reshape(-1),
            np.meshgrid(lon, lat)[1].reshape(-1),
        )
        reshape = True
    utm_x, utm_y = projection(lon, lat)
    if reshape:
        utm_x, utm_y = (
            utm_x.reshape(*reshape_size)[0, :],
            utm_y.reshape(*reshape_size)[:, 0],
        )

    # Return the UTM coordinates
    return utm_x, utm_y


def spatial_gradient(data: xr.DataArray) -> xr.DataArray:
    """
    Calculate spatial gradient of a DataArray with dimensions (time, latitude, longitude).

    Parameters
    ----------
    data : xr.DataArray
        Input data with dimensions (time, latitude, longitude).

    Returns
    -------
    xr.DataArray
        Gradient magnitude with same dimensions as input.

    Notes
    -----
    The gradient is calculated using central differences, accounting for
    latitude-dependent grid spacing in spherical coordinates.
    """

    # Initialize gradient array
    var_grad = xr.zeros_like(data)

    # Get latitude values in radians for spherical coordinate correction
    lat_rad = np.pi * np.abs(data.latitude.values) / 180.0

    # Calculate gradients using vectorized operations
    for t in range(len(data.time)):
        var_val = data.isel(time=t).values

        # calculate gradient (matrix)
        m_c = var_val[1:-1, 1:-1]
        m_l = np.roll(var_val, -1, axis=1)[1:-1, 1:-1]
        m_r = np.roll(var_val, +1, axis=1)[1:-1, 1:-1]
        m_u = np.roll(var_val, -1, axis=0)[1:-1, 1:-1]
        m_d = np.roll(var_val, +1, axis=0)[1:-1, 1:-1]
        m_phi = lat_rad[1:-1]

        dpx1 = (m_c - m_l) / np.cos(m_phi[:, None])
        dpx2 = (m_r - m_c) / np.cos(m_phi[:, None])
        dpy1 = m_c - m_d
        dpy2 = m_u - m_c

        vg = (dpx1**2 + dpx2**2) / 2 + (dpy1**2 + dpy2**2) / 2
        var_grad[t, 1:-1, 1:-1] = vg

    # Set attributes
    var_grad.attrs["units"] = "m^2/s^2"
    var_grad.attrs["name"] = "Gradient"

    return var_grad



from .._base_wrappers import BaseModelWrapper


class SfincsModelWrapper(BaseModelWrapper):
    """
    Wrapper for the SFINCS model.

    Attributes
    ----------
    default_parameters : dict
        The default parameters type for the wrapper.
    available_launchers : dict
        The available launchers for the wrapper.
    """

    default_parameters = {}

    available_launchers = {
        "docker": "docker run --rm -v .:/case_dir -w /case_dir deltares/sfincs-cpu",
        "cluster": "launchSfincs.sh",
    }

    def __init__(
        self,
        templates_dir: str,
        model_parameters: dict,
        output_dir: str,
        templates_name: dict = "all",
        debug: bool = True,
    ) -> None:
        """
        Initialize the SFINCS model wrapper.
        """

        super().__init__(
            templates_dir=templates_dir,
            model_parameters=model_parameters,
            output_dir=output_dir,
            templates_name=templates_name,
            default_parameters=self.default_parameters,
        )
        self.set_logger_name(
            name=self.__class__.__name__, level="DEBUG" if debug else "INFO"
        )



import cmocean
import matplotlib.colors as mcolors
import numpy as np
from matplotlib import cm
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap

default_colors = [
    "#636EFA",
    "#EF553B",
    "#00CC96",
    "#AB63FA",
    "#FFA15A",
    "#19D3F3",
    "#FF6692",
    "#B6E880",
    "#FF97FF",
    "#FECB52",
]


scatter_defaults = {
    "figsize": (9, 8),
    "marker": ".",
    "color_data": "#00CC96",
    "color_subset": "#AB63FA",
    "alpha_data": 0.5,
    "alpha_subset": 0.7,
    "size_data": 10,
    "size_centroid": 70,
    "fontsize": 12,
}


def get_config_variables():
    config_variables = {}

    config_variables["geo500hpa"] = {
        "cmap": "Blues",
        "limits": (45000, 60000),
        "label": "Geo500Hpa [m]",
    }

    config_variables["mslp"] = {
        "cmap": "RdBu_r",
        "limits": (1014 - 20, 1014 + 20),
        "label": "MSLP [mbar]",
    }

    config_variables["mslp_grad"] = {
        "cmap": "BuGn",
        "limits": (0, 100),
        "label": "MSLP - grad [mbar]",
    }

    config_variables["sst"] = {
        "cmap": "hot_r",
        "limits": (0, 30),
        "label": "SST - [°C]",
    }

    config_variables["other"] = {
        "cmap": "rainbow",
        "limits": (None, None),
        "label": " ",
    }

    return config_variables


def colors_awt():
    # 6 AWT colors
    l_colors_dwt = [
        (155 / 255.0, 0, 0),
        (1, 0, 0),
        (255 / 255.0, 216 / 255.0, 181 / 255.0),
        (164 / 255.0, 226 / 255.0, 231 / 255.0),
        (0 / 255.0, 190 / 255.0, 255 / 255.0),
        (51 / 255.0, 0 / 255.0, 207 / 255.0),
    ]

    return np.array(l_colors_dwt)


def colors_mjo():
    "custom colors for MJO 25 categories"

    l_named_colors = [
        "lightskyblue",
        "deepskyblue",
        "royalblue",
        "mediumblue",
        "darkblue",
        "darkblue",
        "darkturquoise",
        "turquoise",
        "maroon",
        "saddlebrown",
        "chocolate",
        "gold",
        "orange",
        "orangered",
        "red",
        "firebrick",
        "Purple",
        "darkorchid",
        "mediumorchid",
        "magenta",
        "mediumslateblue",
        "blueviolet",
        "darkslateblue",
        "indigo",
        "darkgray",
    ]

    # get rgb colors as numpy array
    np_colors_rgb = np.array([mcolors.to_rgb(c) for c in l_named_colors])

    return np_colors_rgb


def colors_dwt(num_clusters):
    # 42 DWT colors
    l_colors_dwt = [
        (1.0000, 0.1344, 0.0021),
        (1.0000, 0.2669, 0.0022),
        (1.0000, 0.5317, 0.0024),
        (1.0000, 0.6641, 0.0025),
        (1.0000, 0.9287, 0.0028),
        (0.9430, 1.0000, 0.0029),
        (0.6785, 1.0000, 0.0031),
        (0.5463, 1.0000, 0.0032),
        (0.2821, 1.0000, 0.0035),
        (0.1500, 1.0000, 0.0036),
        (0.0038, 1.0000, 0.1217),
        (0.0039, 1.0000, 0.2539),
        (0.0039, 1.0000, 0.4901),
        (0.0039, 1.0000, 0.6082),
        (0.0039, 1.0000, 0.8444),
        (0.0039, 1.0000, 0.9625),
        (0.0039, 0.8052, 1.0000),
        (0.0039, 0.6872, 1.0000),
        (0.0040, 0.4510, 1.0000),
        (0.0040, 0.3329, 1.0000),
        (0.0040, 0.0967, 1.0000),
        (0.1474, 0.0040, 1.0000),
        (0.2655, 0.0040, 1.0000),
        (0.5017, 0.0040, 1.0000),
        (0.6198, 0.0040, 1.0000),
        (0.7965, 0.0040, 1.0000),
        (0.8848, 0.0040, 1.0000),
        (1.0000, 0.0040, 0.9424),
        (1.0000, 0.0040, 0.8541),
        (1.0000, 0.0040, 0.6774),
        (1.0000, 0.0040, 0.5890),
        (1.0000, 0.0040, 0.4124),
        (1.0000, 0.0040, 0.3240),
        (1.0000, 0.0040, 0.1473),
        (0.9190, 0.1564, 0.2476),
        (0.7529, 0.3782, 0.4051),
        (0.6699, 0.4477, 0.4584),
        (0.5200, 0.5200, 0.5200),
        (0.4595, 0.4595, 0.4595),
        (0.4100, 0.4100, 0.4100),
        (0.3706, 0.3706, 0.3706),
        (0.2000, 0.2000, 0.2000),
        (0, 0, 0),
    ]

    # get first N colors
    np_colors_base = np.array(l_colors_dwt)
    np_colors_rgb = np_colors_base[:num_clusters]

    return np_colors_rgb


def colors_fams_3():
    "custom colors for 3 waves families"

    l_named_colors = [
        "gold",
        "darkgreen",
        "royalblue",
    ]

    # get rgb colors as numpy array
    np_colors_rgb = np.array([mcolors.to_rgb(c) for c in l_named_colors])

    return np_colors_rgb


def colors_interp(num_clusters):
    # generate spectral colormap
    scm = cm.get_cmap("Spectral", num_clusters)

    # use normalize
    mnorm = mcolors.Normalize(vmin=0, vmax=num_clusters)

    # interpolate colors from cmap
    l_colors = []
    for i in range(num_clusters):
        l_colors.append(scm(mnorm(i)))

    # return numpy array
    np_colors_rgb = np.array(l_colors)[:, :-1]

    return np_colors_rgb


def get_cluster_colors(num_clusters):
    "Choose colors or Interpolate custom colormap to number of clusters"

    if num_clusters == 6:
        np_colors_rgb = colors_awt()  # Annual Weather Types

    if num_clusters == 25:
        np_colors_rgb = colors_mjo()  # MJO Categories

    elif num_clusters in [36, 42]:
        np_colors_rgb = colors_dwt(num_clusters)  # Daily Weather Types

    else:
        np_colors_rgb = colors_interp(num_clusters)  # interpolate

    return np_colors_rgb


def GetFamsColors(num_fams):
    "Choose colors or Interpolate custom colormap to number of waves families"

    if num_fams == 3:
        np_colors_rgb = colors_fams_3()  # choosen colors

    else:
        np_colors_rgb = colors_interp(num_fams)  # interpolate

    return np_colors_rgb


def colormap_bathy(topat, topag):  # maximum topo, minimum bati
    """
    returns custom colormap for bathymetry plot
    """

    colors2 = "YlGnBu_r"
    colors1 = cmocean.cm.turbid

    bottom = plt.get_cmap(colors2, -topag * 100)
    top = plt.get_cmap(colors1, topat * 100)

    newcolors = np.vstack(
        (
            bottom(np.linspace(0, 0.8, -topag * 100)),
            top(np.linspace(0.1, 1, topat * 100)),
        )
    )

    return ListedColormap(newcolors)


def colormap_spectra():
    top = cm.get_cmap("RdBu", 128)
    bottom = cm.get_cmap("rainbow", 128)
    newcolors = np.vstack(
        (top(np.linspace(0.5, 0.8, 50)), bottom(np.linspace(0.2, 1, 128)))
    )
    newcmp = ListedColormap(newcolors, name="newcmp")

    return newcmp



import os

import numpy as np

from bluemath_tk.datamining.lhs import LHS
from bluemath_tk.datamining.mda import MDA
from bluemath_tk.waves.series import series_TMA
from bluemath_tk.wrappers.swash.swash_wrapper import SwashModelWrapper


class VeggySwashModelWrapper(SwashModelWrapper):
    """
    Wrapper for the SWASH model with vegetation.
    """

    def build_case(
        self,
        case_context: dict,
        case_dir: str,
    ) -> None:
        """
        Build the input files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        # Build the input waves
        waves_dict = {
            "H": case_context["Hs"],
            "T": np.sqrt(
                (case_context["Hs"] * 2 * np.pi) / (9.806 * case_context["Hs_L0"])
            ),
            "gamma": 2,
            "warmup": 180,
            "deltat": 1,
            "tendc": 1800,
        }
        waves = series_TMA(waves=waves_dict, depth=10.0)
        # Save the waves to a file
        self.write_array_in_file(
            array=waves, filename=os.path.join(case_dir, "waves.bnd")
        )

    def build_cases(
        self,
        mode: str = "one_by_one",
    ) -> None:
        """
        Build the input files for all cases.

        Parameters
        ----------
        mode : str, optional
            The mode to build the cases. Default is "one_by_one".

        Raises
        ------
        ValueError
            If the cases were not properly built
        """

        super().build_cases(mode=mode)
        if not self.cases_context or not self.cases_dirs:
            raise ValueError("Cases were not properly built.")
        for case_context, case_dir in zip(self.cases_context, self.cases_dirs):
            self.build_case(
                case_context=case_context,
                case_dir=case_dir,
            )


# Usage example
if __name__ == "__main__":
    # Define the input parameters
    templates_dir = (
        "/home/tausiaj/GitHub-GeoOcean/BlueMath/bluemath_tk/wrappers/swash/templates/"
    )
    # Get 5 cases using LHS and MDA
    lhs = LHS(num_dimensions=3)
    lhs_data = lhs.generate(
        dimensions_names=["Hs", "Hs_L0", "vegetation_height"],
        lower_bounds=[0.5, 0.0, 0.0],
        upper_bounds=[3.0, 0.05, 1.5],
        num_samples=500,
    )
    mda = MDA(num_centers=5)
    mda.logger.setLevel("DEBUG")
    mda.fit(data=lhs_data)
    model_parameters = mda.centroids.to_dict(orient="list")
    output_dir = "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_cases/swash/"
    # Create an instance of the SWASH model wrapper
    swash_wrapper = VeggySwashModelWrapper(
        templates_dir=templates_dir,
        model_parameters=model_parameters,
        output_dir=output_dir,
    )
    # Build the input files
    swash_wrapper.build_cases(mode="one_by_one")
    # List available launchers
    print(swash_wrapper.list_available_launchers())
    # Run the model using docker_serial launcher
    # swash_wrapper.run_cases(launcher="docker_serial", num_workers=5)
    # Post-process the output files
    postprocessed_data = swash_wrapper.postprocess_cases(num_workers=1)
    print(postprocessed_data)



# --------------------------- ONLY FOR DEVELOPMENT MODE ---------------------------
# This setup.py file is used to configure the setup for the BlueMath project.
# It specifies metadata about the project and its dependencies.
# This file is ONLY useful for development mode, where you can install
# the project in an editable state using the command `pip install -e .`.
# This allows you to make changes to the code and have them immediately reflected
# without needing to reinstall the package.
# To install the package, change name of pyproject.toml to pyproject.toml.bak

from setuptools import setup, find_packages

setup(
    name="BlueMath",
    author="GeoOcean Group",
    author_email="bluemath@unican.es",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/GeoOcean/BlueMath_tk",
    packages=find_packages(),  # Automatically find packages in the directory
    include_package_data=True,
    install_requires=[
        "numpy",
        "pandas",
        "xarray",
        "netcdf4",
        "dask",
        "distributed",
        "zarr",
        "scipy",
        "scikit-learn",
        "matplotlib",
        "plotly",
        "cartopy",
        "cdsapi",
        "jinja2",
        "requests",
        "aiohttp",
        "minisom",
        "statsmodels",
        "regionmask",
        "wavespectra",
        "cmocean",
        "hydromt-sfincs",
    ],
    classifiers=["Programming Language :: Python :: 3.11"],
    python_requires=">=3.11",  # Specify the Python version required
)



import os
import re
from typing import List

import pandas as pd
import scipy.io as sio
import xarray as xr

from .._base_wrappers import BaseModelWrapper


class SwanModelWrapper(BaseModelWrapper):
    """
    Wrapper for the SWAN model.
    https://swanmodel.sourceforge.io/online_doc/swanuse/swanuse.html

    Attributes
    ----------
    default_parameters : dict
        The default parameters type for the wrapper.
    available_launchers : dict
        The available launchers for the wrapper.
    """

    default_parameters = {
        "hs": float,
        "tp": float,
        "dir": float,
        "spr": float,
    }

    available_launchers = {
        "bash": "swanrun -input input",
        "docker_serial": "docker run --rm -v .:/case_dir -w /case_dir geoocean/rocky8 swan_serial.exe",
    }

    output_variables = {
        "Depth": {
            "long_name": "Water depth at the point",
            "units": "m",
        },
        "Hsig": {
            "long_name": "Significant wave height",
            "units": "m",
        },
        "Tm02": {
            "long_name": "Mean wave period",
            "units": "s",
        },
        "Dir": {
            "long_name": "Wave direction",
            "units": "degrees",
        },
        "PkDir": {
            "long_name": "Peak wave direction",
            "units": "degrees",
        },
        "TPsmoo": {
            "long_name": "Peak wave period",
            "units": "s",
        },
        "Dspr": {
            "long_name": "Directional spread",
            "units": "degrees",
        },
    }

    def __init__(
        self,
        templates_dir: str,
        model_parameters: dict,
        output_dir: str,
        templates_name: dict = "all",
        debug: bool = True,
    ) -> None:
        """
        Initialize the SWAN model wrapper.
        """

        super().__init__(
            templates_dir=templates_dir,
            model_parameters=model_parameters,
            output_dir=output_dir,
            templates_name=templates_name,
            default_parameters=self.default_parameters,
        )
        self.set_logger_name(
            name=self.__class__.__name__, level="DEBUG" if debug else "INFO"
        )

    def list_available_output_variables(self) -> List[str]:
        """
        List available output variables.

        Returns
        -------
        List[str]
            The available output variables.
        """

        return list(self.output_variables.keys())

    def _convert_case_output_files_to_nc(
        self, case_num: int, output_path: str, output_vars: List[str]
    ) -> xr.Dataset:
        """
        Convert mat file to netCDF file.

        Parameters
        ----------
        case_num : int
            The case number.
        output_path : str
            The output path.
        output_vars : List[str]
            The output variables to use.

        Returns
        -------
        xr.Dataset
            The xarray Dataset.
        """

        # Read mat file
        output_dict = sio.loadmat(output_path)

        # Create Dataset
        ds_output_dict = {var: (("Yp", "Xp"), output_dict[var]) for var in output_vars}
        ds = xr.Dataset(
            ds_output_dict,
            coords={"Xp": output_dict["Xp"][0, :], "Yp": output_dict["Yp"][:, 0]},
        )

        # assign correct coordinate case_num
        ds.coords["case_num"] = case_num

        return ds

    def get_case_percentage_from_file(self, output_log_file: str) -> str:
        """
        Get the case percentage from the output log file.

        Parameters
        ----------
        output_log_file : str
            The output log file.

        Returns
        -------
        str
            The case percentage.
        """

        if not os.path.exists(output_log_file):
            return "0 %"

        progress_pattern = r"OK in\s+(\d+\.\d+)\s*%"
        with open(output_log_file, "r") as f:
            for line in reversed(f.readlines()):
                match = re.search(progress_pattern, line)
                if match:
                    return f"{match.group(1)} %"

        return "0 %"  # if no progress is found

    def monitor_cases(self) -> pd.DataFrame:
        """
        Monitor the cases and log relevant information.

        Returns
        -------
        pd.DataFrame
            The cases percentage.
        """

        cases_percentage = {}

        for case_dir in self.cases_dirs:
            output_log_file = os.path.join(case_dir, "wrapper_out.log")
            progress = self.get_case_percentage_from_file(
                output_log_file=output_log_file
            )
            cases_percentage[os.path.basename(case_dir)] = progress

        return pd.DataFrame(cases_percentage.items(), columns=["Case", "Percentage"])

    def postprocess_case(
        self,
        case_num: int,
        case_dir: str,
        output_vars: List[str] = ["Hsig", "Tm02", "Dir"],
    ) -> xr.Dataset:
        """
        Convert mat ouput files to netCDF file.

        Parameters
        ----------
        case_num : int
            The case number.
        case_dir : str
            The case directory.
        output_vars : list, optional
            The output variables to postprocess. Default is None.

        Returns
        -------
        xr.Dataset
            The postprocessed Dataset.
        """

        if output_vars is None:
            self.logger.info("Postprocessing all available variables.")
            output_vars = list(self.output_variables.keys())

        output_nc_path = os.path.join(case_dir, "output.nc")
        if not os.path.exists(output_nc_path):
            # Convert tab files to netCDF file
            output_path = os.path.join(case_dir, "output.mat")
            output_nc = self._convert_case_output_files_to_nc(
                case_num=case_num,
                output_path=output_path,
                output_vars=output_vars,
            )
            output_nc.to_netcdf(os.path.join(case_dir, "output.nc"))
        else:
            self.logger.info("Reading existing output.nc file.")
            output_nc = xr.open_dataset(output_nc_path)

        return output_nc

    def join_postprocessed_files(
        self, postprocessed_files: List[xr.Dataset]
    ) -> xr.Dataset:
        """
        Join postprocessed files in a single Dataset.

        Parameters
        ----------
        postprocessed_files : list
            The postprocessed files.

        Returns
        -------
        xr.Dataset
            The joined Dataset.
        """

        return xr.concat(postprocessed_files, dim="case_num")



from typing import List, Tuple

import numpy as np
import pandas as pd
import xarray as xr
from dask.diagnostics.progress import ProgressBar
from matplotlib import cm
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap
from wavespectra.input.swan import read_swan

from ..core.dask import setup_dask_client
from ..core.plotting.base_plotting import DefaultStaticPlotting


def generate_swan_cases(
    frequencies_array: np.ndarray,
    directions_array: np.ndarray,
) -> xr.Dataset:
    """
    Generate the SWAN cases monocromatic wave parameters.

    Parameters
    ----------
    directions_array : np.ndarray
        The directions array.
    frequencies_array : np.ndarray
        The frequencies array.

    Returns
    -------
    xr.Dataset
        The SWAN monocromatic cases Dataset with coordinates freq and dir.
    """

    # Wave parameters
    gamma = 50  # waves gamma
    spr = 2  # waves directional spread

    # Initialize data arrays for each variable
    hs = np.zeros((len(directions_array), len(frequencies_array)))
    tp = np.zeros((len(directions_array), len(frequencies_array)))
    gamma_arr = np.full((len(directions_array), len(frequencies_array)), gamma)
    spr_arr = np.full((len(directions_array), len(frequencies_array)), spr)

    # Fill hs and tp arrays
    for i, freq in enumerate(frequencies_array):
        period = 1 / freq
        hs_val = 1.0 if period > 5 else 0.1
        hs[:, i] = hs_val
        tp[:, i] = np.round(period, 4)

    # Create xarray Dataset
    ds = xr.Dataset(
        data_vars={
            "hs": (["dir", "freq"], hs),
            "tp": (["dir", "freq"], tp),
            "spr": (["dir", "freq"], spr_arr),
            "gamma": (["dir", "freq"], gamma_arr),
        },
        coords={
            "dir": directions_array,
            "freq": frequencies_array,
        },
    )

    # To get DataFrame if needed:
    # df = ds.to_dataframe().reset_index()

    return ds


def process_kp_coefficients(
    list_of_input_spectra: List[str],
    list_of_output_spectra: List[str],
) -> xr.Dataset:
    """
    Process the kp coefficients from the output and input spectra.

    Parameters
    ----------
    list_of_input_spectra : List[str]
        The list of input spectra files.
    list_of_output_spectra : List[str]
        The list of output spectra files.

    Returns
    -------
    xr.Dataset
        The kp coefficients Dataset in frequency and direction.
    """

    output_kp_list = []

    for i, (input_spec_file, output_spec_file) in enumerate(
        zip(list_of_input_spectra, list_of_output_spectra)
    ):
        try:
            input_spec = read_swan(input_spec_file).squeeze().efth
            output_spec = (
                read_swan(output_spec_file)
                .efth.squeeze()
                .drop_vars("time")
                .expand_dims({"case_num": [i]})
            )
            kp = output_spec / input_spec.sum(dim=["freq", "dir"])
            output_kp_list.append(kp)
        except Exception as e:
            print(f"Error processing {input_spec_file} and {output_spec_file}")
            print(e)

    return (
        xr.concat(output_kp_list, dim="case_num")
        .fillna(0.0)
        .sortby("freq")
        .sortby("dir")
    )


def reconstruc_spectra(
    offshore_spectra: xr.Dataset,
    kp_coeffs: xr.Dataset,
):
    """
    Reconstruct the onshore spectra using offshore spectra and kp coefficients.

    Parameters
    ----------
    offshore_spectra : xr.Dataset
        The offshore spectra dataset.
    kp_coeffs : xr.Dataset
        The kp coefficients dataset.

    Returns
    -------
    xr.Dataset
        The reconstructed onshore spectra dataset.
    """

    # Setup Dask client
    client = setup_dask_client(n_workers=4, memory_limit=0.2)

    try:
        # Process with controlled chunks
        offshore_spectra_chunked = offshore_spectra.chunk({"time": 24 * 7})
        kp_coeffs_chunked = kp_coeffs.chunk({"site": 1})
        with ProgressBar():
            onshore_spectra = (
                (offshore_spectra_chunked * kp_coeffs_chunked).sum(dim="case_num")
            ).compute()
        return onshore_spectra

    finally:
        client.close()


def plot_selected_subset_parameters(
    selected_subset: pd.DataFrame,
    color: str = "blue",
    **kwargs,
) -> Tuple[plt.figure, plt.axes]:
    """
    Plot the selected subset parameters.

    Parameters
    ----------
    selected_subset : pd.DataFrame
        The selected subset parameters.
    color : str, optional
        The color to use in the plot. Default is "blue".
    **kwargs : dict, optional
        Additional keyword arguments to be passed to the scatter plot function.

    Returns
    -------
    plt.figure
        The figure object containing the plot.
    plt.axes
        Array of axes objects for the subplots.
    """

    # Create figure and axes
    default_static_plot = DefaultStaticPlotting()
    fig, axes = default_static_plot.get_subplots(
        nrows=len(selected_subset) - 1,
        ncols=len(selected_subset) - 1,
        sharex=False,
        sharey=False,
    )

    for c1, v1 in enumerate(list(selected_subset.columns)[1:]):
        for c2, v2 in enumerate(list(selected_subset.columns)[:-1]):
            default_static_plot.plot_scatter(
                ax=axes[c2, c1],
                x=selected_subset[v1],
                y=selected_subset[v2],
                c=color,
                alpha=0.6,
                **kwargs,
            )
            if c1 == c2:
                axes[c2, c1].set_xlabel(list(selected_subset.columns)[c1 + 1])
                axes[c2, c1].set_ylabel(list(selected_subset.columns)[c2])
            elif c1 > c2:
                axes[c2, c1].xaxis.set_ticklabels([])
                axes[c2, c1].yaxis.set_ticklabels([])
            else:
                fig.delaxes(axes[c2, c1])

    return fig, axes


def plot_selected_cases_grid(
    frequencies: np.ndarray,
    directions: np.ndarray,
    figsize: Tuple[int, int] = (8, 8),
    **kwargs,
):
    """
    Plot the selected subset parameters.

    Parameters
    ----------
    frequencies : np.ndarray
        The frequencies array.
    directions : np.ndarray
        The directions array.
    figsize : tuple, optional
        The figure size. Default is (8, 8).
    **kwargs : dict, optional
        Additional keyword arguments to be passed to the pcolormesh function.
    """

    # generate figure and axes
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(1, 1, 1, projection="polar")

    # prepare data
    x = np.append(np.deg2rad(directions - 7.5), np.deg2rad(directions - 7.5)[0])
    y = np.append(0, frequencies)
    z = (
        np.array(range(len(frequencies) * len(directions)))
        .reshape(len(directions), len(frequencies))
        .T
    )

    # custom colormap
    cmn = np.vstack(
        (
            cm.get_cmap("plasma", 124)(np.linspace(0, 0.9, 70)),
            cm.get_cmap("magma_r", 124)(np.linspace(0.1, 0.4, 80)),
            cm.get_cmap("rainbow_r", 124)(np.linspace(0.1, 0.8, 80)),
            cm.get_cmap("Blues_r", 124)(np.linspace(0.4, 0.8, 40)),
            cm.get_cmap("cubehelix_r", 124)(np.linspace(0.1, 0.8, 80)),
        )
    )
    cmn = ListedColormap(cmn, name="cmn")

    # plot cases id
    p1 = ax.pcolormesh(
        x,
        y,
        z,
        vmin=0,
        vmax=np.nanmax(z),
        edgecolor="grey",
        linewidth=0.005,
        cmap=cmn,
        shading="flat",
        **kwargs,
    )

    # customize axes
    ax.set_theta_zero_location("N", offset=0)
    ax.set_theta_direction(-1)
    ax.tick_params(
        axis="both",
        colors="black",
        labelsize=14,
        pad=10,
    )

    # add colorbar
    plt.colorbar(p1, pad=0.1, shrink=0.7).set_label("Case ID", fontsize=16)



import unittest
import numpy as np
import pandas as pd
from bluemath_tk.datamining.kma import KMA


class TestKMA(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(
            {
                "Hs": np.random.rand(1000) * 7,
                "Tp": np.random.rand(1000) * 20,
                "Dir": np.random.rand(1000) * 360,
            }
        )
        self.kma = KMA(num_clusters=10)

    def test_fit(self):
        self.kma.fit(data=self.df, min_number_of_points=50)
        self.assertIsInstance(self.kma.centroids, pd.DataFrame)
        self.assertEqual(self.kma.centroids.shape[0], 10)

    def test_predict(self):
        data_sample = pd.DataFrame(
            {
                "Hs": np.random.rand(15) * 7,
                "Tp": np.random.rand(15) * 20,
                "Dir": np.random.rand(15) * 360,
            }
        )
        self.kma.fit(data=self.df)
        nearest_centroids, nearest_centroid_df = self.kma.predict(data=data_sample)
        self.assertIsInstance(nearest_centroids, np.ndarray)
        self.assertEqual(len(nearest_centroids), 15)
        self.assertIsInstance(nearest_centroid_df, pd.DataFrame)
        self.assertEqual(nearest_centroid_df.shape[0], 15)

    def test_fit_predict(self):
        predicted_labels, predicted_labels_df = self.kma.fit_predict(
            data=self.df, min_number_of_points=50
        )
        _unique_labels, counts = np.unique(predicted_labels, return_counts=True)
        self.assertTrue(np.all(counts >= 50))
        self.assertIsInstance(predicted_labels, np.ndarray)
        self.assertEqual(len(predicted_labels), 1000)
        self.assertIsInstance(predicted_labels_df, pd.DataFrame)
        self.assertEqual(predicted_labels_df.shape[0], 1000)


if __name__ == "__main__":
    unittest.main()
